from pomagma.compiler import extensional
from pomagma.compiler.expressions import Expression
from pomagma.compiler.expressions import Expression_0
from pomagma.compiler.expressions import Expression_2
from pomagma.compiler.sequents import Sequent
from pomagma.compiler.util import inputs

assert extensional  # pacify pyflakes


A = Expression_0('A')
Y = Expression_0('Y')
APP = Expression_2('APP')
EQUAL = Expression_2('EQUAL')


@inputs(Expression)
def desugar_expr(self):
    expr = Expression.make(self.name, *map(desugar_expr, self.args))
    if expr.name == 'FUN':
        var, body = expr.args
        assert var.is_var(), var
        expr = body.abstract(var)
    elif expr.name == 'FIX':
        var, body = expr.args
        assert var.is_var(), var
        expr = APP(Y, body.abstract(var))
    elif expr.name == 'ABIND':
        s, r, body = expr.args
        assert s.is_var(), s
        assert r.is_var(), r
        expr = APP(A, body.abstract(r).abstract(s))
    elif expr.name == 'FIXES':
        typ, inhab = expr.args
        expr = EQUAL(APP(typ, inhab), inhab)
    return expr


@inputs(Sequent)
def desugar_sequent(self):
    return Sequent(
        set(desugar_expr(a) for a in self.antecedents),
        set(desugar_expr(s) for s in self.succedents))


@inputs(dict)
def desugar_theory(theory):
    rules = map(desugar_sequent, theory['rules'])
    facts = []
    for fact in map(desugar_expr, theory['facts']):
        if fact.vars:
            rules.append(Sequent(set(), set([fact])))
        else:
            facts.append(fact)
    return {'facts': facts, 'rules': rules}



import os
from parsable import parsable
import pomagma.atlas
import pomagma.util


@parsable
def info(filename):
    '''
    Print info about a structure file.
    '''
    pomagma.atlas.print_info(filename)


@parsable
def cp(theory, source, destin):
    '''
    Copy and recompress structure file.
    '''
    assert source != destin
    with pomagma.util.mutex(source):
        assert os.path.exists(source)
        assert not os.path.exists(destin)
        with pomagma.cartographer.load(theory, source) as db:
            db.dump(destin)


@parsable
def blobs(root):
    '''
    Find all blobs referenced under a root directory.
    '''
    for blob in pomagma.atlas.find_used_blobs(root):
        print blob


parsable()



import cProfile as profile
import contextlib
import glob
import multiprocessing
import os
from parsable import parsable
import pstats
import re
import simplejson as json
from pomagma.compiler import compiler
from pomagma.compiler import completion
from pomagma.compiler import extensional
from pomagma.compiler import frontend
from pomagma.compiler import parser
from pomagma.compiler import sequencer
from pomagma.compiler.compiler import compile_full
from pomagma.compiler.compiler import compile_given
from pomagma.compiler.compiler import get_events
from pomagma.compiler.plans import add_costs
from pomagma.compiler.sugar import desugar_expr
from pomagma.compiler.sugar import desugar_theory
from pomagma.compiler.util import find_theories

SRC = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
ROOT = os.path.dirname(SRC)


def up_to_date(infiles, outfiles):
    if not all(os.path.exists(f) for f in infiles + outfiles):
        return False
    infiles = glob.glob(os.path.join(SRC, 'compiler', '*.py')) + infiles
    intimes = map(os.path.getmtime, infiles)
    outtimes = map(os.path.getmtime, outfiles)
    return max(intimes) < min(outtimes)


def load_theory(filename):
    theory = parser.parse_theory_file(filename)
    theory = desugar_theory(theory)
    theory['rules'].sort()
    theory['facts'].sort()
    return theory


def json_load(filename):
    with open(filename) as f:
        return json.load(f)


def parse_bool(arg):
    return {'true': True, 'false': False}[arg.lower()]


@parsable
def abstract(*args):
    '''
    Abstract variables from expression.
    Examples:
        abstract x 'APP x y'
        abstract x y z 'APP APP x z APP y z'
    '''
    args = map(parser.parse_string_to_expr, args)
    expression, vars = args[-1], args[:-1]
    for var in reversed(vars):
        expression = expression.abstract(var)
    print expression


@parsable
def desugar(*exprs):
    '''
    Convert lambda terms to combinators.
    Examples:
        desugar FUN x APP x y
        desugar FUN x FUN y FUN z APP APP x z APP y z
    '''
    for expr in map(parser.parse_string_to_expr, exprs):
        print expr
        print '  =', desugar_expr(expr)


@parsable
def complete(*facts):
    '''
    Complete a set of facts.
    '''
    facts = set(map(parser.parse_string_to_expr, facts))
    facts = completion.complete(facts)
    facts = sorted(facts)
    for fact in facts:
        print fact


@contextlib.contextmanager
def writer(outfile=None):
    if outfile is None:

        def write(line):
            print line
            print

        yield write
    else:
        with open(outfile, 'w') as f:

            def write(line):
                f.write(line)
                f.write('\n\n')

            yield write


def print_compiles(compiles):
    for cost, seq, plan in compiles:
        print '# cost = {0}'.format(cost)
        print '# infer {0}'.format(seq)
        print re.sub(': ', '\n', repr(plan))
        print


def measure_sequent(sequent):
    print '-' * 78
    print 'Compiling full search: {0}'.format(sequent)
    compiles = compile_full(sequent)
    print_compiles(compiles)
    full_cost = add_costs(c for (c, _, _) in compiles)

    incremental_cost = None
    for event in get_events(sequent):
        print 'Compiling incremental search given: {0}'.format(event)
        compiles = compile_given(sequent, event)
        print_compiles(compiles)
        if event.args:
            cost = add_costs(c for (c, _, _) in compiles)
            if incremental_cost:
                incremental_cost = add_costs([incremental_cost, cost])
            else:
                incremental_cost = cost
        else:
            pass  # event is only triggered once, so ignore cost

    print '# full cost =', full_cost, 'incremental cost =', incremental_cost


@parsable
def contrapositves(*filenames):
    '''
    Close rules under contrapositve.
    '''
    if not filenames:
        filenames = find_theories()
    sequents = []
    for filename in filenames:
        sequents += load_theory(filename)['rules']
    for sequent in sequents:
        print sequent.ascii()
        print
        for neg in sequent.contrapositives():
            print neg.ascii(indent=4)
            print


@parsable
def normalize(*filenames):
    '''
    Show normalized rule set derived from each rule.
    '''
    if not filenames:
        filenames = find_theories()
    sequents = []
    for filename in filenames:
        sequents += load_theory(filename)['rules']
    for sequent in sequents:
        print sequent.ascii()
        print
        for neg in sequent.contrapositives():
            print neg.ascii(indent=4)
            print


@parsable
def extract_tasks(infile, outfile=None):
    '''
    Extract tasks from facts and rules, but do not compile to programs.
    '''
    theory = load_theory(infile)
    with writer(outfile) as write:
        facts = theory['facts']
        for sequent in theory['rules']:
            facts += extensional.derive_facts(sequent)
        facts.sort()
        write('\n'.join(map(str, facts)))
        for sequent in theory['rules']:
            write(sequent.ascii())
            for normal in sorted(compiler.normalize(sequent)):
                write(normal.ascii(indent=4))
            for event in sorted(compiler.get_events(sequent)):
                write('    Given: {}'.format(event))
                for normal in sorted(compiler.normalize_given(sequent, event)):
                    write(normal.ascii(indent=8))


def _extract_tasks((infile, outfile)):
    extract_tasks(infile, outfile)


@parsable
def batch_extract_tasks(*filenames, **kwargs):
    '''
    Extract tasks from infiles '*.theory', saving to '*.tasks'.
    Options: parallel=true
    '''
    if not filenames:
        filenames = find_theories()
    pairs = []
    for infile in filenames:
        infile = os.path.abspath(infile)
        assert infile.endswith('.theory'), infile
        outfile = infile.replace('.theory', '.tasks')
        if not up_to_date([infile], [outfile]):
            pairs.append((infile, outfile))
    parallel = parse_bool(kwargs.get('parallel', 'true'))
    map_ = multiprocessing.Pool().map if parallel else map
    map_(_extract_tasks, pairs)


@parsable
def measure(*filenames):
    '''
    Measure complexity of rules in files.
    '''
    if not filenames:
        filenames = find_theories()
    sequents = []
    for filename in filenames:
        sequents += load_theory(filename)['rules']
    for sequent in sequents:
        measure_sequent(sequent)


@parsable
def test_compile(*filenames):
    '''
    Compile rules -> programs for the virtual machine.
    '''
    if not filenames:
        filenames = find_theories()
    for stem_rules in filenames:
        programs = ['# {filename}'.format(filename=stem_rules)]
        assert stem_rules[-len('.theory'):] == '.theory', stem_rules

        sequents = load_theory(stem_rules)['rules']
        for sequent in sequents:
            for cost, seq, plan in compile_full(sequent):
                programs += [
                    '',
                    '# using {}'.format(sequent),
                    '# infer '.format(seq),
                    '# cost = '.format(cost),
                ]
                plan.program(programs)

            for event in get_events(sequent):
                for cost, seq, plan in compile_given(sequent, event):
                    programs += [
                        '',
                        '# given {}'.format(event),
                        '# using {}'.format(sequent),
                        '# infer {}'.format(seq),
                        '# cost {}'.format(cost),
                    ]
                    plan.program(programs)

        print '\n'.join(programs)


@parsable
def profile_tasks(*filenames, **kwargs):
    '''
    Profile task generation (first part of compiler chain).
    Optional keyword arguments:
        loadfrom = None
        saveto = 'tasks.pstats'
    '''
    if not filenames:
        filenames = find_theories()
    loadfrom = kwargs.get('loadfrom')
    saveto = kwargs.get('saveto', 'tasks.pstats')
    if loadfrom is None:
        command = 'batch_extract_tasks({}, parallel=false)'.format(
            ', '.join(map('"{}"'.format, filenames)))
        print 'profiling {}'.format(command)
        profile.run(command, saveto)
        loadfrom = saveto
    stats = pstats.Stats(loadfrom)
    stats.strip_dirs()
    line_count = 50
    for sortby in ['time']:
        stats.sort_stats(sortby)
        stats.print_stats(line_count)


@parsable
def profile_compile(*filenames, **kwargs):
    '''
    Profile full compiler chain (task generation + optimization).
    Optional keyword arguments:
        loadfrom = None
        saveto = 'compile.pstats'
    '''
    if not filenames:
        filenames = find_theories()
    loadfrom = kwargs.get('loadfrom')
    saveto = kwargs.get('saveto', 'compile.pstats')
    if loadfrom is None:
        command = 'compile({}, frontend_out="/dev/null")'.format(
            ', '.join(map('"{}"'.format, filenames)))
        print 'profiling {}'.format(command)
        profile.run(command, saveto)
        loadfrom = saveto
    stats = pstats.Stats(loadfrom)
    stats.strip_dirs()
    line_count = 50
    for sortby in ['time']:
        stats.sort_stats(sortby)
        stats.print_stats(line_count)


@parsable
def test_close_rules(infile, is_extensional=True):
    '''
    Compile extensionally some.theory -> some.derived.facts.
    '''
    assert infile.endswith('.theory')
    rules = load_theory(infile)['rules']
    for rule in rules:
        print
        print '#', rule
        for fact in extensional.derive_facts(rule):
            print fact
            if is_extensional:
                extensional.validate(fact)


def relpath(string):
    is_path = '.' in string and '/' in string  # heuristic
    if is_path:
        return os.path.relpath(string, ROOT)
    else:
        return string


@parsable
def compile(*infiles, **kwargs):
    '''
    Compile rules -> programs for the virtual machine.
    Optional keyword arguments:
        symbols_out=$POMAGMA_ROOT/src/theory/<STEM>.symbols
        facts_out=$POMAGMA_ROOT/src/theory/<STEM>.facts
        programs_out=$POMAGMA_ROOT/src/theory/<STEM>.programs
        optimized_out=$POMAGMA_ROOT/src/theory/<STEM>.optimized.programs
        extensional=true
    '''
    stem = infiles[-1].split('.')[0]
    symbols_out = kwargs.get(
        'symbols_out',
        os.path.join(SRC, 'theory', '{0}.symbols'.format(stem)))
    facts_out = kwargs.get(
        'facts_out',
        os.path.join(SRC, 'theory', '{0}.facts'.format(stem)))
    programs_out = kwargs.get(
        'programs_out',
        os.path.join(SRC, 'theory', '{0}.programs'.format(stem)))
    optimized_out = kwargs.get(
        'optimized_out',
        os.path.join(SRC, 'theory', '{0}.optimized.programs'.format(stem)))
    is_extensional = parse_bool(kwargs.get('extensional', 'true'))

    argstring = ' '.join(
        [relpath(path) for path in infiles] +
        [
            '{0}={1}'.format(key, relpath(path))
            for key, path in kwargs.iteritems()
        ])
    header = (
        '# This file was auto generated by pomagma using:\n'
        '# python -m pomagma.compiler compile {0}'.format(argstring)
    )

    rules = []
    facts = []
    for infile in infiles:
        theory = load_theory(infile)
        rules += theory['rules']
        facts += theory['facts']
    if is_extensional:
        for rule in rules:
            facts += extensional.derive_facts(rule)
    rules.sort()
    facts.sort()

    symbols = frontend.write_symbols(rules, facts)
    with open(symbols_out, 'w') as f:
        print '# writing', symbols_out
        f.write(header)
        for arity, name in symbols:
            f.write('\n{} {}'.format(arity, name))

    with open(facts_out, 'w') as f:
        print '# writing', facts_out
        f.write(header)
        for fact in facts:
            assert fact.is_rel(), 'bad fact: %s' % fact
            f.write('\n')
            f.write(fact.polish)

    programs = frontend.write_programs(rules)
    with open(programs_out, 'w') as f:
        print '# writing', programs_out
        f.write(header)
        for line in programs:
            f.write('\n')
            f.write(line)

    lines = sequencer.load_lines(programs_out)
    optimized = sequencer.optimize(lines)
    with open(optimized_out, 'w') as f:
        print '# writing', optimized_out
        f.write(header)
        for line in optimized:
            f.write('\n')
            f.write(line)


def _compile(param):
    compile(*param['args'], **param['kwargs'])


@parsable
def batch_compile(parallel=True):
    '''
    Compile all theories in parallel.
    '''
    params = []
    theories_json = os.path.join(SRC, 'theory', 'theories.json')
    theories = json_load(theories_json)
    for name, spec in theories.iteritems():
        infiles = sorted(
            os.path.join(SRC, 'theory', '{}.theory'.format(t))
            for t in spec['theories']
        )
        symbols_out = os.path.join(SRC, 'theory', '{}.symbols'.format(name))
        facts_out = os.path.join(SRC, 'theory', '{}.facts'.format(name))
        programs_out = os.path.join(SRC, 'theory', '{}.programs'.format(name))
        optimized_out = os.path.join(
            SRC,
            'theory',
            '{}.optimized.programs'.format(name))
        outfiles = [symbols_out, facts_out, programs_out]
        if not up_to_date(infiles + [theories_json], outfiles):
            params.append({
                'args': infiles,
                'kwargs': {
                    'symbols_out': symbols_out,
                    'facts_out': facts_out,
                    'programs_out': programs_out,
                    'optimized_out': optimized_out,
                    'extensional': str(spec.get('extensional', True)),
                }})
    map_ = multiprocessing.Pool().map if parallel else map
    map_(_compile, params)


if __name__ == '__main__':
    parsable()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: messages.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)




DESCRIPTOR = _descriptor.FileDescriptor(
  name='messages.proto',
  package='pomagma.protobuf',
  serialized_pb='\n\x0emessages.proto\x12\x10pomagma.protobuf\"\xf0\x08\n\x13\x43\x61rtographerRequest\x12\x38\n\x04\x63rop\x18\n \x01(\x0b\x32*.pomagma.protobuf.CartographerRequest.Crop\x12>\n\x07\x64\x65\x63lare\x18\x0b \x01(\x0b\x32-.pomagma.protobuf.CartographerRequest.Declare\x12<\n\x06\x61ssume\x18\x0c \x01(\x0b\x32,.pomagma.protobuf.CartographerRequest.Assume\x12:\n\x05infer\x18\r \x01(\x0b\x32+.pomagma.protobuf.CartographerRequest.Infer\x12>\n\x07\x65xecute\x18\x0e \x01(\x0b\x32-.pomagma.protobuf.CartographerRequest.Execute\x12\x42\n\taggregate\x18\x14 \x01(\x0b\x32/.pomagma.protobuf.CartographerRequest.Aggregate\x12@\n\x08validate\x18\x1e \x01(\x0b\x32..pomagma.protobuf.CartographerRequest.Validate\x12\x38\n\x04info\x18( \x01(\x0b\x32*.pomagma.protobuf.CartographerRequest.Info\x12\x38\n\x04\x64ump\x18) \x01(\x0b\x32*.pomagma.protobuf.CartographerRequest.Dump\x12\x38\n\x04trim\x18* \x03(\x0b\x32*.pomagma.protobuf.CartographerRequest.Trim\x12\x44\n\nconjecture\x18+ \x01(\x0b\x32\x30.pomagma.protobuf.CartographerRequest.Conjecture\x12\x38\n\x04stop\x18\x63 \x01(\x0b\x32*.pomagma.protobuf.CartographerRequest.Stop\x1a\x1b\n\x04\x43rop\x12\x13\n\x08headroom\x18\x01 \x01(\r:\x01\x30\x1a$\n\x07\x44\x65\x63lare\x12\x19\n\x11nullary_functions\x18\x01 \x03(\t\x1a\x1a\n\x06\x41ssume\x12\x10\n\x08\x66\x61\x63ts_in\x18\x01 \x02(\t\x1a\x19\n\x05Infer\x12\x10\n\x08priority\x18\x01 \x02(\r\x1a\x1a\n\x07\x45xecute\x12\x0f\n\x07program\x18\x01 \x02(\t\x1a\x1e\n\tAggregate\x12\x11\n\tsurvey_in\x18\x01 \x02(\t\x1a\n\n\x08Validate\x1a\x06\n\x04Info\x1a\x19\n\x04\x44ump\x12\x11\n\tworld_out\x18\x01 \x02(\t\x1a;\n\x04Trim\x12\x0c\n\x04size\x18\x01 \x02(\r\x12\x13\n\x0btemperature\x18\x02 \x02(\x08\x12\x10\n\x08\x66ilename\x18\x03 \x02(\t\x1aG\n\nConjecture\x12\x13\n\x0b\x64iverge_out\x18\x01 \x02(\t\x12\x11\n\tequal_out\x18\x02 \x02(\t\x12\x11\n\tmax_count\x18\x03 \x02(\x04\x1a\x06\n\x04Stop\"\xab\x08\n\x14\x43\x61rtographerResponse\x12\x39\n\x04\x63rop\x18\n \x01(\x0b\x32+.pomagma.protobuf.CartographerResponse.Crop\x12?\n\x07\x64\x65\x63lare\x18\x0b \x01(\x0b\x32..pomagma.protobuf.CartographerResponse.Declare\x12=\n\x06\x61ssume\x18\x0c \x01(\x0b\x32-.pomagma.protobuf.CartographerResponse.Assume\x12;\n\x05infer\x18\r \x01(\x0b\x32,.pomagma.protobuf.CartographerResponse.Infer\x12?\n\x07\x65xecute\x18\x0e \x01(\x0b\x32..pomagma.protobuf.CartographerResponse.Execute\x12\x43\n\taggregate\x18\x14 \x01(\x0b\x32\x30.pomagma.protobuf.CartographerResponse.Aggregate\x12\x41\n\x08validate\x18\x1e \x01(\x0b\x32/.pomagma.protobuf.CartographerResponse.Validate\x12\x39\n\x04info\x18( \x01(\x0b\x32+.pomagma.protobuf.CartographerResponse.Info\x12\x39\n\x04\x64ump\x18) \x01(\x0b\x32+.pomagma.protobuf.CartographerResponse.Dump\x12\x39\n\x04trim\x18* \x03(\x0b\x32+.pomagma.protobuf.CartographerResponse.Trim\x12\x45\n\nconjecture\x18+ \x01(\x0b\x32\x31.pomagma.protobuf.CartographerResponse.Conjecture\x12\x39\n\x04stop\x18\x63 \x01(\x0b\x32+.pomagma.protobuf.CartographerResponse.Stop\x1a\x06\n\x04\x43rop\x1a\t\n\x07\x44\x65\x63lare\x1aZ\n\x06\x41ssume\x12\x11\n\tpos_count\x18\x01 \x02(\x04\x12\x11\n\tneg_count\x18\x02 \x02(\x04\x12\x13\n\x0bmerge_count\x18\x03 \x02(\x04\x12\x15\n\rignored_count\x18\x04 \x02(\x04\x1a\x1e\n\x05Infer\x12\x15\n\rtheorem_count\x18\x01 \x02(\x04\x1a\t\n\x07\x45xecute\x1a\x0b\n\tAggregate\x1a\n\n\x08Validate\x1a\x1a\n\x04Info\x12\x12\n\nitem_count\x18\x01 \x02(\r\x1a\x06\n\x04\x44ump\x1a\x06\n\x04Trim\x1a\x38\n\nConjecture\x12\x15\n\rdiverge_count\x18\x01 \x02(\x04\x12\x13\n\x0b\x65qual_count\x18\x02 \x02(\x04\x1a\x06\n\x04Stop')




_CARTOGRAPHERREQUEST_CROP = _descriptor.Descriptor(
  name='Crop',
  full_name='pomagma.protobuf.CartographerRequest.Crop',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='headroom', full_name='pomagma.protobuf.CartographerRequest.Crop.headroom', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=804,
  serialized_end=831,
)

_CARTOGRAPHERREQUEST_DECLARE = _descriptor.Descriptor(
  name='Declare',
  full_name='pomagma.protobuf.CartographerRequest.Declare',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nullary_functions', full_name='pomagma.protobuf.CartographerRequest.Declare.nullary_functions', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=833,
  serialized_end=869,
)

_CARTOGRAPHERREQUEST_ASSUME = _descriptor.Descriptor(
  name='Assume',
  full_name='pomagma.protobuf.CartographerRequest.Assume',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='facts_in', full_name='pomagma.protobuf.CartographerRequest.Assume.facts_in', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=871,
  serialized_end=897,
)

_CARTOGRAPHERREQUEST_INFER = _descriptor.Descriptor(
  name='Infer',
  full_name='pomagma.protobuf.CartographerRequest.Infer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='priority', full_name='pomagma.protobuf.CartographerRequest.Infer.priority', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=899,
  serialized_end=924,
)

_CARTOGRAPHERREQUEST_EXECUTE = _descriptor.Descriptor(
  name='Execute',
  full_name='pomagma.protobuf.CartographerRequest.Execute',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='program', full_name='pomagma.protobuf.CartographerRequest.Execute.program', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=926,
  serialized_end=952,
)

_CARTOGRAPHERREQUEST_AGGREGATE = _descriptor.Descriptor(
  name='Aggregate',
  full_name='pomagma.protobuf.CartographerRequest.Aggregate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='survey_in', full_name='pomagma.protobuf.CartographerRequest.Aggregate.survey_in', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=954,
  serialized_end=984,
)

_CARTOGRAPHERREQUEST_VALIDATE = _descriptor.Descriptor(
  name='Validate',
  full_name='pomagma.protobuf.CartographerRequest.Validate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=986,
  serialized_end=996,
)

_CARTOGRAPHERREQUEST_INFO = _descriptor.Descriptor(
  name='Info',
  full_name='pomagma.protobuf.CartographerRequest.Info',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=998,
  serialized_end=1004,
)

_CARTOGRAPHERREQUEST_DUMP = _descriptor.Descriptor(
  name='Dump',
  full_name='pomagma.protobuf.CartographerRequest.Dump',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='world_out', full_name='pomagma.protobuf.CartographerRequest.Dump.world_out', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1006,
  serialized_end=1031,
)

_CARTOGRAPHERREQUEST_TRIM = _descriptor.Descriptor(
  name='Trim',
  full_name='pomagma.protobuf.CartographerRequest.Trim',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='size', full_name='pomagma.protobuf.CartographerRequest.Trim.size', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='temperature', full_name='pomagma.protobuf.CartographerRequest.Trim.temperature', index=1,
      number=2, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='filename', full_name='pomagma.protobuf.CartographerRequest.Trim.filename', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1033,
  serialized_end=1092,
)

_CARTOGRAPHERREQUEST_CONJECTURE = _descriptor.Descriptor(
  name='Conjecture',
  full_name='pomagma.protobuf.CartographerRequest.Conjecture',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='diverge_out', full_name='pomagma.protobuf.CartographerRequest.Conjecture.diverge_out', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='equal_out', full_name='pomagma.protobuf.CartographerRequest.Conjecture.equal_out', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_count', full_name='pomagma.protobuf.CartographerRequest.Conjecture.max_count', index=2,
      number=3, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1094,
  serialized_end=1165,
)

_CARTOGRAPHERREQUEST_STOP = _descriptor.Descriptor(
  name='Stop',
  full_name='pomagma.protobuf.CartographerRequest.Stop',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1167,
  serialized_end=1173,
)

_CARTOGRAPHERREQUEST = _descriptor.Descriptor(
  name='CartographerRequest',
  full_name='pomagma.protobuf.CartographerRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='crop', full_name='pomagma.protobuf.CartographerRequest.crop', index=0,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='declare', full_name='pomagma.protobuf.CartographerRequest.declare', index=1,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='assume', full_name='pomagma.protobuf.CartographerRequest.assume', index=2,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='infer', full_name='pomagma.protobuf.CartographerRequest.infer', index=3,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='execute', full_name='pomagma.protobuf.CartographerRequest.execute', index=4,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='aggregate', full_name='pomagma.protobuf.CartographerRequest.aggregate', index=5,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='validate', full_name='pomagma.protobuf.CartographerRequest.validate', index=6,
      number=30, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='info', full_name='pomagma.protobuf.CartographerRequest.info', index=7,
      number=40, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dump', full_name='pomagma.protobuf.CartographerRequest.dump', index=8,
      number=41, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='trim', full_name='pomagma.protobuf.CartographerRequest.trim', index=9,
      number=42, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conjecture', full_name='pomagma.protobuf.CartographerRequest.conjecture', index=10,
      number=43, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stop', full_name='pomagma.protobuf.CartographerRequest.stop', index=11,
      number=99, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_CARTOGRAPHERREQUEST_CROP, _CARTOGRAPHERREQUEST_DECLARE, _CARTOGRAPHERREQUEST_ASSUME, _CARTOGRAPHERREQUEST_INFER, _CARTOGRAPHERREQUEST_EXECUTE, _CARTOGRAPHERREQUEST_AGGREGATE, _CARTOGRAPHERREQUEST_VALIDATE, _CARTOGRAPHERREQUEST_INFO, _CARTOGRAPHERREQUEST_DUMP, _CARTOGRAPHERREQUEST_TRIM, _CARTOGRAPHERREQUEST_CONJECTURE, _CARTOGRAPHERREQUEST_STOP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=37,
  serialized_end=1173,
)


_CARTOGRAPHERRESPONSE_CROP = _descriptor.Descriptor(
  name='Crop',
  full_name='pomagma.protobuf.CartographerResponse.Crop',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=804,
  serialized_end=810,
)

_CARTOGRAPHERRESPONSE_DECLARE = _descriptor.Descriptor(
  name='Declare',
  full_name='pomagma.protobuf.CartographerResponse.Declare',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=833,
  serialized_end=842,
)

_CARTOGRAPHERRESPONSE_ASSUME = _descriptor.Descriptor(
  name='Assume',
  full_name='pomagma.protobuf.CartographerResponse.Assume',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pos_count', full_name='pomagma.protobuf.CartographerResponse.Assume.pos_count', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='neg_count', full_name='pomagma.protobuf.CartographerResponse.Assume.neg_count', index=1,
      number=2, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='merge_count', full_name='pomagma.protobuf.CartographerResponse.Assume.merge_count', index=2,
      number=3, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ignored_count', full_name='pomagma.protobuf.CartographerResponse.Assume.ignored_count', index=3,
      number=4, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1975,
  serialized_end=2065,
)

_CARTOGRAPHERRESPONSE_INFER = _descriptor.Descriptor(
  name='Infer',
  full_name='pomagma.protobuf.CartographerResponse.Infer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='theorem_count', full_name='pomagma.protobuf.CartographerResponse.Infer.theorem_count', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=2067,
  serialized_end=2097,
)

_CARTOGRAPHERRESPONSE_EXECUTE = _descriptor.Descriptor(
  name='Execute',
  full_name='pomagma.protobuf.CartographerResponse.Execute',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=926,
  serialized_end=935,
)

_CARTOGRAPHERRESPONSE_AGGREGATE = _descriptor.Descriptor(
  name='Aggregate',
  full_name='pomagma.protobuf.CartographerResponse.Aggregate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=954,
  serialized_end=965,
)

_CARTOGRAPHERRESPONSE_VALIDATE = _descriptor.Descriptor(
  name='Validate',
  full_name='pomagma.protobuf.CartographerResponse.Validate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=986,
  serialized_end=996,
)

_CARTOGRAPHERRESPONSE_INFO = _descriptor.Descriptor(
  name='Info',
  full_name='pomagma.protobuf.CartographerResponse.Info',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='item_count', full_name='pomagma.protobuf.CartographerResponse.Info.item_count', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=2135,
  serialized_end=2161,
)

_CARTOGRAPHERRESPONSE_DUMP = _descriptor.Descriptor(
  name='Dump',
  full_name='pomagma.protobuf.CartographerResponse.Dump',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1006,
  serialized_end=1012,
)

_CARTOGRAPHERRESPONSE_TRIM = _descriptor.Descriptor(
  name='Trim',
  full_name='pomagma.protobuf.CartographerResponse.Trim',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1033,
  serialized_end=1039,
)

_CARTOGRAPHERRESPONSE_CONJECTURE = _descriptor.Descriptor(
  name='Conjecture',
  full_name='pomagma.protobuf.CartographerResponse.Conjecture',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='diverge_count', full_name='pomagma.protobuf.CartographerResponse.Conjecture.diverge_count', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='equal_count', full_name='pomagma.protobuf.CartographerResponse.Conjecture.equal_count', index=1,
      number=2, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=2179,
  serialized_end=2235,
)

_CARTOGRAPHERRESPONSE_STOP = _descriptor.Descriptor(
  name='Stop',
  full_name='pomagma.protobuf.CartographerResponse.Stop',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1167,
  serialized_end=1173,
)

_CARTOGRAPHERRESPONSE = _descriptor.Descriptor(
  name='CartographerResponse',
  full_name='pomagma.protobuf.CartographerResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='crop', full_name='pomagma.protobuf.CartographerResponse.crop', index=0,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='declare', full_name='pomagma.protobuf.CartographerResponse.declare', index=1,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='assume', full_name='pomagma.protobuf.CartographerResponse.assume', index=2,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='infer', full_name='pomagma.protobuf.CartographerResponse.infer', index=3,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='execute', full_name='pomagma.protobuf.CartographerResponse.execute', index=4,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='aggregate', full_name='pomagma.protobuf.CartographerResponse.aggregate', index=5,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='validate', full_name='pomagma.protobuf.CartographerResponse.validate', index=6,
      number=30, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='info', full_name='pomagma.protobuf.CartographerResponse.info', index=7,
      number=40, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dump', full_name='pomagma.protobuf.CartographerResponse.dump', index=8,
      number=41, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='trim', full_name='pomagma.protobuf.CartographerResponse.trim', index=9,
      number=42, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conjecture', full_name='pomagma.protobuf.CartographerResponse.conjecture', index=10,
      number=43, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stop', full_name='pomagma.protobuf.CartographerResponse.stop', index=11,
      number=99, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_CARTOGRAPHERRESPONSE_CROP, _CARTOGRAPHERRESPONSE_DECLARE, _CARTOGRAPHERRESPONSE_ASSUME, _CARTOGRAPHERRESPONSE_INFER, _CARTOGRAPHERRESPONSE_EXECUTE, _CARTOGRAPHERRESPONSE_AGGREGATE, _CARTOGRAPHERRESPONSE_VALIDATE, _CARTOGRAPHERRESPONSE_INFO, _CARTOGRAPHERRESPONSE_DUMP, _CARTOGRAPHERRESPONSE_TRIM, _CARTOGRAPHERRESPONSE_CONJECTURE, _CARTOGRAPHERRESPONSE_STOP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1176,
  serialized_end=2243,
)

_CARTOGRAPHERREQUEST_CROP.containing_type = _CARTOGRAPHERREQUEST;
_CARTOGRAPHERREQUEST_DECLARE.containing_type = _CARTOGRAPHERREQUEST;
_CARTOGRAPHERREQUEST_ASSUME.containing_type = _CARTOGRAPHERREQUEST;
_CARTOGRAPHERREQUEST_INFER.containing_type = _CARTOGRAPHERREQUEST;
_CARTOGRAPHERREQUEST_EXECUTE.containing_type = _CARTOGRAPHERREQUEST;
_CARTOGRAPHERREQUEST_AGGREGATE.containing_type = _CARTOGRAPHERREQUEST;
_CARTOGRAPHERREQUEST_VALIDATE.containing_type = _CARTOGRAPHERREQUEST;
_CARTOGRAPHERREQUEST_INFO.containing_type = _CARTOGRAPHERREQUEST;
_CARTOGRAPHERREQUEST_DUMP.containing_type = _CARTOGRAPHERREQUEST;
_CARTOGRAPHERREQUEST_TRIM.containing_type = _CARTOGRAPHERREQUEST;
_CARTOGRAPHERREQUEST_CONJECTURE.containing_type = _CARTOGRAPHERREQUEST;
_CARTOGRAPHERREQUEST_STOP.containing_type = _CARTOGRAPHERREQUEST;
_CARTOGRAPHERREQUEST.fields_by_name['crop'].message_type = _CARTOGRAPHERREQUEST_CROP
_CARTOGRAPHERREQUEST.fields_by_name['declare'].message_type = _CARTOGRAPHERREQUEST_DECLARE
_CARTOGRAPHERREQUEST.fields_by_name['assume'].message_type = _CARTOGRAPHERREQUEST_ASSUME
_CARTOGRAPHERREQUEST.fields_by_name['infer'].message_type = _CARTOGRAPHERREQUEST_INFER
_CARTOGRAPHERREQUEST.fields_by_name['execute'].message_type = _CARTOGRAPHERREQUEST_EXECUTE
_CARTOGRAPHERREQUEST.fields_by_name['aggregate'].message_type = _CARTOGRAPHERREQUEST_AGGREGATE
_CARTOGRAPHERREQUEST.fields_by_name['validate'].message_type = _CARTOGRAPHERREQUEST_VALIDATE
_CARTOGRAPHERREQUEST.fields_by_name['info'].message_type = _CARTOGRAPHERREQUEST_INFO
_CARTOGRAPHERREQUEST.fields_by_name['dump'].message_type = _CARTOGRAPHERREQUEST_DUMP
_CARTOGRAPHERREQUEST.fields_by_name['trim'].message_type = _CARTOGRAPHERREQUEST_TRIM
_CARTOGRAPHERREQUEST.fields_by_name['conjecture'].message_type = _CARTOGRAPHERREQUEST_CONJECTURE
_CARTOGRAPHERREQUEST.fields_by_name['stop'].message_type = _CARTOGRAPHERREQUEST_STOP
_CARTOGRAPHERRESPONSE_CROP.containing_type = _CARTOGRAPHERRESPONSE;
_CARTOGRAPHERRESPONSE_DECLARE.containing_type = _CARTOGRAPHERRESPONSE;
_CARTOGRAPHERRESPONSE_ASSUME.containing_type = _CARTOGRAPHERRESPONSE;
_CARTOGRAPHERRESPONSE_INFER.containing_type = _CARTOGRAPHERRESPONSE;
_CARTOGRAPHERRESPONSE_EXECUTE.containing_type = _CARTOGRAPHERRESPONSE;
_CARTOGRAPHERRESPONSE_AGGREGATE.containing_type = _CARTOGRAPHERRESPONSE;
_CARTOGRAPHERRESPONSE_VALIDATE.containing_type = _CARTOGRAPHERRESPONSE;
_CARTOGRAPHERRESPONSE_INFO.containing_type = _CARTOGRAPHERRESPONSE;
_CARTOGRAPHERRESPONSE_DUMP.containing_type = _CARTOGRAPHERRESPONSE;
_CARTOGRAPHERRESPONSE_TRIM.containing_type = _CARTOGRAPHERRESPONSE;
_CARTOGRAPHERRESPONSE_CONJECTURE.containing_type = _CARTOGRAPHERRESPONSE;
_CARTOGRAPHERRESPONSE_STOP.containing_type = _CARTOGRAPHERRESPONSE;
_CARTOGRAPHERRESPONSE.fields_by_name['crop'].message_type = _CARTOGRAPHERRESPONSE_CROP
_CARTOGRAPHERRESPONSE.fields_by_name['declare'].message_type = _CARTOGRAPHERRESPONSE_DECLARE
_CARTOGRAPHERRESPONSE.fields_by_name['assume'].message_type = _CARTOGRAPHERRESPONSE_ASSUME
_CARTOGRAPHERRESPONSE.fields_by_name['infer'].message_type = _CARTOGRAPHERRESPONSE_INFER
_CARTOGRAPHERRESPONSE.fields_by_name['execute'].message_type = _CARTOGRAPHERRESPONSE_EXECUTE
_CARTOGRAPHERRESPONSE.fields_by_name['aggregate'].message_type = _CARTOGRAPHERRESPONSE_AGGREGATE
_CARTOGRAPHERRESPONSE.fields_by_name['validate'].message_type = _CARTOGRAPHERRESPONSE_VALIDATE
_CARTOGRAPHERRESPONSE.fields_by_name['info'].message_type = _CARTOGRAPHERRESPONSE_INFO
_CARTOGRAPHERRESPONSE.fields_by_name['dump'].message_type = _CARTOGRAPHERRESPONSE_DUMP
_CARTOGRAPHERRESPONSE.fields_by_name['trim'].message_type = _CARTOGRAPHERRESPONSE_TRIM
_CARTOGRAPHERRESPONSE.fields_by_name['conjecture'].message_type = _CARTOGRAPHERRESPONSE_CONJECTURE
_CARTOGRAPHERRESPONSE.fields_by_name['stop'].message_type = _CARTOGRAPHERRESPONSE_STOP
DESCRIPTOR.message_types_by_name['CartographerRequest'] = _CARTOGRAPHERREQUEST
DESCRIPTOR.message_types_by_name['CartographerResponse'] = _CARTOGRAPHERRESPONSE

class CartographerRequest(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType

  class Crop(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERREQUEST_CROP

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest.Crop)

  class Declare(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERREQUEST_DECLARE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest.Declare)

  class Assume(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERREQUEST_ASSUME

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest.Assume)

  class Infer(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERREQUEST_INFER

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest.Infer)

  class Execute(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERREQUEST_EXECUTE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest.Execute)

  class Aggregate(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERREQUEST_AGGREGATE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest.Aggregate)

  class Validate(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERREQUEST_VALIDATE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest.Validate)

  class Info(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERREQUEST_INFO

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest.Info)

  class Dump(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERREQUEST_DUMP

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest.Dump)

  class Trim(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERREQUEST_TRIM

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest.Trim)

  class Conjecture(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERREQUEST_CONJECTURE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest.Conjecture)

  class Stop(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERREQUEST_STOP

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest.Stop)
  DESCRIPTOR = _CARTOGRAPHERREQUEST

  # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerRequest)

class CartographerResponse(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType

  class Crop(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERRESPONSE_CROP

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse.Crop)

  class Declare(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERRESPONSE_DECLARE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse.Declare)

  class Assume(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERRESPONSE_ASSUME

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse.Assume)

  class Infer(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERRESPONSE_INFER

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse.Infer)

  class Execute(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERRESPONSE_EXECUTE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse.Execute)

  class Aggregate(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERRESPONSE_AGGREGATE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse.Aggregate)

  class Validate(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERRESPONSE_VALIDATE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse.Validate)

  class Info(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERRESPONSE_INFO

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse.Info)

  class Dump(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERRESPONSE_DUMP

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse.Dump)

  class Trim(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERRESPONSE_TRIM

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse.Trim)

  class Conjecture(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERRESPONSE_CONJECTURE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse.Conjecture)

  class Stop(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CARTOGRAPHERRESPONSE_STOP

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse.Stop)
  DESCRIPTOR = _CARTOGRAPHERRESPONSE

  # @@protoc_insertion_point(class_scope:pomagma.protobuf.CartographerResponse)


# @@protoc_insertion_point(module_scope)



from nose.tools import assert_equal
from pomagma.compiler.parser import parse_string_to_expr
from pomagma.compiler.sugar import desugar_expr
from pomagma.util.testing import for_each


EXAMPLES = [map(parse_string_to_expr, e) for e in [
    ('x', 'x'),
    ('FUN x x', 'I'),
    ('FUN x APP f x', 'f'),
    ('FIX x APP f x', 'APP Y f'),
    ('ABIND s r COMP r s', 'APP A CB'),
    ('FIXES BOOL I', 'EQUAL APP BOOL I I'),
]]


@for_each(EXAMPLES)
def test_desugar((expr, expected)):
    actual = desugar_expr(expr)
    assert_equal(actual, expected)



import os


def create_directories(path):
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except os.error as e:
            if not os.path.exists(path):
                raise e


def creat(filename, mode):
    assert isinstance(mode, int), mode
    # this may be insufficient to set permissions
    # see http://stackoverflow.com/a/15015748/1224437
    fid = os.open(filename, os.O_CREAT | os.O_WRONLY | os.O_TRUNC, mode)
    return os.fdopen(fid, 'wb')



import os
import pomagma.util
import re
import zmq
from google.protobuf.descriptor import FieldDescriptor
from pomagma.cartographer import messages_pb2 as messages


CONTEXT = zmq.Context()
POLL_TIMEOUT_MS = 1000
Request = messages.CartographerRequest
Response = messages.CartographerResponse


class Client(object):

    def __init__(self, address, poll_callback):
        assert isinstance(address, basestring), address
        assert callable(poll_callback), poll_callback
        self._poll_callback = poll_callback
        self._socket = CONTEXT.socket(zmq.REQ)
        print 'connecting to cartographer at', address
        self._socket.connect(address)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._socket.close()

    def _call(self, request):
        raw_request = request.SerializeToString()
        self._socket.send(raw_request, 0)
        while not self._socket.poll(timeout=POLL_TIMEOUT_MS):
            self._poll_callback()
        raw_reply = self._socket.recv(0)
        reply = Response()
        reply.ParseFromString(raw_reply)
        for field in Request.DESCRIPTOR.fields:
            if field.label == FieldDescriptor.LABEL_OPTIONAL:
                expected = request.HasField(field.name)
                actual = reply.HasField(field.name)
                assert actual == expected, field.name
            elif field.label == FieldDescriptor.LABEL_REPEATED:
                expected = len(getattr(request, field.name))
                actual = len(getattr(reply, field.name))
                assert actual == expected, field.name
        return reply

    def ping(self):
        request = Request()
        self._call(request)

    def crop(self, headroom=0):
        assert isinstance(headroom, (int, long)), headroom
        assert headroom >= 0, headroom
        request = Request()
        request.crop.SetInParent()
        request.crop.headroom = headroom
        self._call(request)

    def _aggregate(self, survey_in):
        request = Request()
        request.aggregate.SetInParent()
        request.aggregate.survey_in = survey_in
        self._call(request)

    def aggregate(self, survey_in):
        assert isinstance(survey_in, basestring), survey_in
        assert os.path.exists(survey_in), survey_in
        self._aggregate(survey_in)

    def declare(self, *nullary_functions):
        for name in nullary_functions:
            assert isinstance(name, basestring), name
        request = Request()
        request.declare.SetInParent()
        request.declare.nullary_functions += nullary_functions
        self._call(request)

    def _assume(self, facts_in):
        request = Request()
        request.assume.SetInParent()
        request.assume.facts_in = facts_in
        reply = self._call(request)
        return {
            'pos': reply.assume.pos_count,
            'neg': reply.assume.neg_count,
            'merge': reply.assume.merge_count,
            'ignored': reply.assume.ignored_count,
        }

    def assume(self, facts_in):
        assert isinstance(facts_in, basestring), facts_in
        assert os.path.exists(facts_in), facts_in
        return self._assume(facts_in)

    def _infer(self, priority):
        request = Request()
        request.infer.SetInParent()
        request.infer.priority = priority
        reply = self._call(request)
        return reply.infer.theorem_count

    def infer(self, priority):
        assert isinstance(priority, int), priority
        assert priority in [0, 1], priority
        return self._infer(priority)

    def normalize(self):
        for priority in [0, 1]:
            while self.infer(priority):
                pass

    def execute(self, program):
        assert isinstance(program, basestring), program
        assert not re.search('FOR_BLOCK', program), 'cannot parallelize'
        request = Request()
        request.execute.program = program
        self._call(request)

    def validate(self):
        request = Request()
        request.validate.SetInParent()
        self._call(request)

    def info(self):
        request = Request()
        request.info.SetInParent()
        reply = self._call(request)
        info = reply.info
        return {'item_count': info.item_count}

    def _dump(self, world_out):
        request = Request()
        request.dump.SetInParent()
        request.dump.world_out = world_out
        self._call(request)

    def dump(self, world_out):
        assert isinstance(world_out, basestring), world_out
        assert os.path.exists(os.path.dirname(os.path.abspath(world_out)))
        with pomagma.util.temp_copy(world_out) as temp_world_out:
            self._dump(temp_world_out)
        assert os.path.exists(world_out), world_out

    def _trim(self, tasks):
        request = Request()
        for task in tasks:
            request_task = request.trim.add()
            request_task.size = task['size']
            request_task.temperature = task['temperature']
            request_task.filename = task['filename']
        self._call(request)

    def trim(self, tasks):
        assert isinstance(tasks, list)
        for task in tasks:
            assert isinstance(task, dict)
        tasks = [task.copy() for task in tasks]
        for task in tasks:
            assert 'size' in task, task
            assert isinstance(task['size'], int), task['size']
            assert 'filename' in task, task
            assert isinstance(task['filename'], basestring), task['filename']
            temperature = task.setdefault('temperature', 1)
            assert temperature in [0, 1], temperature
        filenames = [task['filename'] for task in tasks]
        assert len(filenames) == len(set(filenames)), 'duplicated out file'
        with pomagma.util.temp_copies(filenames) as temp_filenames:
            for task, filename in zip(tasks, temp_filenames):
                task['filename'] = filename
            self._trim(tasks)
        for filename in filenames:
            assert os.path.exists(filename), filename

    def _conjecture(self, diverge_out, equal_out, max_count):
        request = Request()
        request.conjecture.SetInParent()
        request.conjecture.diverge_out = diverge_out
        request.conjecture.equal_out = equal_out
        request.conjecture.max_count = max_count
        reply = self._call(request)
        return {
            'diverge_count': reply.conjecture.diverge_count,
            'equal_count': reply.conjecture.equal_count,
        }

    def conjecture(self, diverge_out, equal_out, max_count=1000):
        assert isinstance(diverge_out, basestring), diverge_out
        assert isinstance(equal_out, basestring), equal_out
        assert isinstance(max_count, int)
        assert max_count > 0
        with pomagma.util.temp_copy(diverge_out) as temp_diverge_out:
            with pomagma.util.temp_copy(equal_out) as temp_equal_out:
                counts = self._conjecture(
                    temp_diverge_out,
                    temp_equal_out,
                    max_count)
        assert os.path.exists(diverge_out), diverge_out
        assert os.path.exists(equal_out), equal_out
        return counts

    def stop(self):
        request = Request()
        request.stop.SetInParent()
        self._call(request)



import os
import contextlib
import pomagma.util
from pomagma.analyst import client
from pomagma.analyst import server


ADDRESS = os.environ.get('POMAGMA_ANALYST_ADDRESS', 'tcp://localhost:34936')


def connect(address=ADDRESS):
    return client.Client(address)


def serve(theory, world, address=ADDRESS, **opts):
    return server.Server(theory, world, address, **opts)


@contextlib.contextmanager
def load(theory, world, **opts):
    address = 'ipc://{}'.format(os.path.join(
        os.path.dirname(pomagma.util.abspath(world)),
        'analyst.socket'))
    with pomagma.util.log_duration():
        server = serve(theory, world, address, **opts)
        with server.connect() as client:
            yield client
        server.stop()



import os
import glob
import functools
import itertools
from itertools import izip
from math import log
from math import exp
import pomagma.util

function = type(lambda x: x)


def DELETE(*args, **kwargs):
    raise ValueError('deleted method')


def logger(message, *args):
    if pomagma.util.LOG_LEVEL >= pomagma.util.LOG_LEVEL_DEBUG:
        print '#', message.format(*args)


class sortedset(set):
    __slots__ = ['_sorted', '_hash']

    def __init__(self, *args, **kwargs):
        set.__init__(self, *args, **kwargs)
        self._sorted = tuple(sorted(set.__iter__(self)))
        self._hash = hash(self._sorted)

    def __iter__(self):
        return iter(self._sorted)

    def __and__(self, other):
        result = set(self)
        result &= other
        return result

    def __or__(self, other):
        result = set(self)
        result |= other
        return result

    def __sub__(self, other):
        result = set(self)
        result -= other
        return result

    def __xor__(self, other):
        result = set(self)
        result ^= other
        return result

    def __hash__(self):
        return self._hash

    # weak immutability
    update = DELETE
    difference_update = DELETE
    intersection_update = DELETE
    symmetric_difference_update = DELETE
    add = DELETE
    remove = DELETE
    discard = DELETE
    pop = DELETE
    __ior__ = DELETE
    __iand__ = DELETE
    __ixor__ = DELETE
    __isub__ = DELETE


def union(sets):
    result = set()
    for s in sets:
        result.update(s)
    return result


def set_with(set_, *elements):
    result = set(set_)
    for e in elements:
        result.add(e)
    return set_.__class__(result)


def set_without(set_, *elements):
    result = set(set_)
    for e in elements:
        result.remove(e)
    return set_.__class__(result)


def log_sum_exp(*args):
    if args:
        shift = max(args)
        return log(sum(exp(arg - shift) for arg in args)) + shift
    else:
        return -float('inf')


def eval_float44(num):
    '''
    8 bit nonnegative floating point = 4 bit significand + 4 bit exponent.
    Gradually increase from 0 to about 1e6 over inputs 0...255
    such that output is monotone increasing and has small relative increase.
    '''
    assert isinstance(num, int) and 0 <= num and num < 256, num
    nibbles = (num % 16, num / 16)
    return (nibbles[0] + 16) * 2 ** nibbles[1] - 16


def eval_float53(num):
    '''
    8 bit nonnegative floating point = 5 bit significand + 3 bit exponent.
    Gradually increase from 0 to about 8e3 over inputs 0...255
    such that output is monotone increasing and has small relative increase.
    '''
    assert isinstance(num, int) and 0 <= num and num < 256, num
    nibbles = (num % 32, num / 32)
    return (nibbles[0] + 32) * 2 ** nibbles[1] - 32


def inputs(*types):
    def deco(fun):
        @functools.wraps(fun)
        def typed(*args, **kwargs):
            for arg, typ in zip(args, types):
                assert isinstance(arg, typ), arg
            return fun(*args, **kwargs)
        return typed
    return deco


def methodof(class_, name=None):
    def deco(fun):
        if name is not None:
            fun.__name__ = name
        setattr(class_, fun.__name__, fun)
    return deco


def find_theories():
    return glob.glob(os.path.join(pomagma.util.THEORY, '*.theory'))


def memoize_arg(fun):
    cache = {}

    @functools.wraps(fun)
    def memoized(arg):
        try:
            return cache[arg]
        except KeyError:
            result = fun(arg)
            cache[arg] = result
            return result

    return memoized


def memoize_args(fun):
    cache = {}

    @functools.wraps(fun)
    def memoized(*args):
        try:
            return cache[args]
        except KeyError:
            result = fun(*args)
            cache[args] = result
            return result

    return memoized


def get_consts(thing):
    if hasattr(thing, 'consts'):
        return thing.consts
    else:
        return union(get_consts(i) for i in thing)


@inputs(dict)
def permute_symbols(perm, thing):
    if not perm:
        return thing
    elif hasattr(thing, 'permute_symbols'):
        return thing.permute_symbols(perm)
    elif hasattr(thing, '__iter__'):
        return thing.__class__(permute_symbols(perm, i) for i in thing)
    elif isinstance(thing, (int, float)):
        return thing
    else:
        raise ValueError('cannot permute_symbols of {}'.format(thing))


def memoize_modulo_renaming_constants(fun):
    cache = {}

    @functools.wraps(fun)
    def cached(*args):
        consts = sorted(c.name for c in get_consts(args))
        result = None
        for permuted_consts in itertools.permutations(consts):
            perm = {i: j for i, j in izip(consts, permuted_consts) if i != j}
            permuted_args = permute_symbols(perm, args)
            try:
                permuted_result = cache[permuted_args]
            except KeyError:
                continue
            logger('{}: using cache via {}', fun.__name__, perm)
            inverse = {j: i for i, j in perm.iteritems()}
            return permute_symbols(inverse, permuted_result)
        logger('{}: compute', fun.__name__)
        result = fun(*args)
        cache[args] = result
        return result

    return cached


def memoize_make(cls):
    cls.make = staticmethod(memoize_args(cls))
    return cls



import os
from nose import SkipTest
from nose.tools import assert_equal, assert_true, assert_false
import mock
import pomagma.io.s3
from pomagma.util import random_uuid, in_temp_dir


TEST_BUCKET = pomagma.io.s3.try_connect_s3('pomagma-test')


def skipped():
    raise SkipTest()


if TEST_BUCKET is None:
    def requires_auth(fun):
        return skipped
else:
    def requires_auth(fun):
        return fun


def load(filename):
    with open(filename) as f:
        return f.read()


def dump(text, filename):
    dirname = os.path.dirname(filename)
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname)
    with open(filename, 'w') as f:
        f.write(text)


def examples(suffix):
    for depth in range(3):
        dirs = [random_uuid() for _ in range(depth)]
        filename = random_uuid() + suffix
        path = apply(os.path.join, dirs + [filename])
        yield path


@requires_auth
def test_s3():
    with mock.patch('pomagma.io.s3.BUCKET', new=TEST_BUCKET):
        for filename in examples('.store_test.test_s3'):
            text = random_uuid()
            with in_temp_dir():
                dump(text, filename)
                print 'putting to s3...',
                pomagma.io.s3.s3_lazy_put(filename)
                print 'done'
            assert_true(pomagma.io.s3.s3_exists(filename))
            with in_temp_dir():
                print 'getting from s3...',
                pomagma.io.s3.s3_lazy_get(filename)
                print 'done'
                assert_true(os.path.exists(filename))
                assert_equal(load(filename), text)
            print 'removing from s3...',
            pomagma.io.s3.s3_remove(filename)
            print 'done'
            assert_false(pomagma.io.s3.s3_exists(filename))


def test_bzip2():
    for filename in examples('.store_test.test_bzip2'):
        text = random_uuid()
        with in_temp_dir():
            dump(text, filename)
            filename_ext = pomagma.io.s3.bzip2(filename)
            assert_true(os.path.exists(filename_ext))
            os.remove(filename)
            pomagma.io.s3.bunzip2(filename_ext)
            assert_true(os.path.exists(filename))
            assert_equal(load(filename), text)


def test_7z():
    for filename in examples('.store_test.test_7z'):
        text = random_uuid()
        with in_temp_dir():
            dump(text, filename)
            filename_ext = pomagma.io.s3.archive_7z(filename)
            assert_true(os.path.exists(filename_ext))
            os.remove(filename)
            pomagma.io.s3.extract_7z(filename_ext)
            assert_true(os.path.exists(filename))
            assert_equal(load(filename), text)


@requires_auth
def test_no_cache():
    with mock.patch('pomagma.io.s3.BUCKET', new=TEST_BUCKET):
        for filename in examples('.store_test.test_no_cache'):
            text = random_uuid()
            with in_temp_dir():
                dump(text, filename)
                pomagma.io.s3.put(filename)
            with in_temp_dir():
                pomagma.io.s3.get(filename)
                assert_true(os.path.exists(filename))
                assert_equal(load(filename), text)
                pomagma.io.s3.remove(filename)


@requires_auth
def test_cache():
    with mock.patch('pomagma.io.s3.BUCKET', new=TEST_BUCKET):
        for filename in examples('.store_test.test_cache'):
            text = random_uuid()
            with in_temp_dir():
                dump(text, filename)
                pomagma.io.s3.put(filename)
                os.remove(filename)
                pomagma.io.s3.get(filename)
                assert_true(os.path.exists(filename))
                assert_equal(load(filename), text)
                pomagma.io.s3.remove(filename)


@requires_auth
def test_stale_cache():
    with mock.patch('pomagma.io.s3.BUCKET', new=TEST_BUCKET):
        for filename in examples('.store_test.test_stale_cache'):
            text = random_uuid()
            stale_text = random_uuid()
            with in_temp_dir():
                dump(text, filename)
                pomagma.io.s3.put(filename)
                dump(stale_text, filename)
                pomagma.io.s3.bzip2(filename)
                pomagma.io.s3.get(filename)
                assert_true(os.path.exists(filename))
                assert_equal(load(filename), text)
                pomagma.io.s3.remove(filename)



from nose.tools import assert_equal, assert_raises
import pomagma.util
from pomagma.theorist.diverge import Converged
from pomagma.theorist.diverge import Diverged
from pomagma.theorist.diverge import I, K, F, B, C, W, S, Y, TOP
from pomagma.theorist.diverge import converge_step
from pomagma.theorist.diverge import iter_terms
from pomagma.theorist.diverge import parse_term
from pomagma.theorist.diverge import print_term
from pomagma.theorist.diverge import try_converge
from pomagma.theorist.diverge import try_prove_diverge
from pomagma.theorist.diverge import may_diverge
from pomagma.theorist.diverge import must_diverge


a, b, c = ('a',), ('b',), ('c',)  # argument lists
x, y, z = 'x', 'y', 'z'  # arguments


STEPS = [
    ((I,), (TOP,)),
    ((I, a,), a),
    ((I, (x,), a,), (x, a,)),

    ((K,), (TOP,)),
    ((K, a,), a),
    ((K, a, b,), a),
    ((K, (x,), b, c,), (x, c,)),

    ((F,), (TOP,)),
    ((F, a,), (TOP,)),
    ((F, a, b,), b),
    ((F, a, (x,), c,), (x, c,)),

    ((B,), (TOP,)),
    ((B, a,), a),
    ((B, (x,), (y,),), (x, (y, (TOP,),),)),
    ((B, (x,), (y,), a,), (x, (y, a,),)),
    ((B, (x,), (y,), a, b,), (x, (y, a,), b,)),

    ((C,), (TOP,)),
    ((C, a,), a),
    ((C, (x,), a,), (x, (TOP,), a,)),
    ((C, (x,), a, b,), (x, b, a,)),
    ((C, (x,), a, b, c,), (x, b, a, c,)),

    ((W,), (TOP,)),
    ((W, (x,),), (x, (TOP,), (TOP,),)),
    ((W, (x,), a,), (x, a, a,)),
    ((W, (x,), a, b,), (x, a, a, b,)),

    ((S,), (TOP,)),
    ((S, (x,),), (x, (TOP,), (TOP,),)),
    ((S, (x,), (y,),), (x, (TOP,), (y, (TOP,),),)),
    ((S, (x,), (y,), a,), (x, a, (y, a,),)),
    ((S, (x,), (y,), a, b,), (x, a, (y, a,), b,)),

    ((Y,), (TOP,)),
    ((Y, (x,),), (x, (Y, (x,),),)),
    ((Y, (x,), a,), (x, (Y, (x,),), a,)),
]


def _test_converge_step(term, expected):
    actual = converge_step(term)
    assert_equal(expected, actual)


def test_converge_step():
    for term, expected in STEPS:
        yield _test_converge_step, term, expected


def test_converge():
    assert_raises(Converged, converge_step, (TOP,))


def test_www_diverges():
    WWW = (W, (W,), (W,),)
    assert_equal(converge_step(WWW), WWW)
    assert_raises(Diverged, try_converge, WWW, 1)


def assert_diverges(string):
    steps = 10
    term = parse_term(string)
    assert_raises(Diverged, try_converge, term, steps)


def test_diverges():
    strings = [
        'APP APP W W W',
        'APP APP W W APP W W',
        'COMP Y CI',
        'COMP Y CB',
        'COMP Y APP S I',
        'COMP Y APP S W',
        'COMP Y APP S S',
    ]
    for string in strings:
        yield assert_diverges, string


def test_try_prove_diverge():
    atoms = [I, K, B, C, W, S]
    max_atom_count = 3
    max_steps = 20
    with pomagma.util.in_temp_dir():
        with open('source.facts', 'w') as f:
            f.write('# test terms')
            for term in iter_terms(atoms, max_atom_count):
                f.write('\n')
                f.write('EQUAL BOT {}'.format(print_term(term)))
        try_prove_diverge(
            'source.facts',
            'unproven.facts',
            'theorems.facts',
            max_steps)
        with open('theorems.facts') as f:
            for line in f:
                line = line.split('#')[0].strip()
                if line.startswith('EQUAL BOT '):
                    term = parse_term(line[len('EQUAL BOT '):])
                    assert_raises(Diverged, try_converge, term, max_steps)
                elif line.startswith('NLESS ') and line.endswith(' BOT'):
                    term = parse_term(line[len('NLESS '): 1 - len(' BOT')])
                    assert_raises(Converged, try_converge, term, max_steps)
                elif line:
                    raise ValueError('Bad line:\n{}'.format(line))


def test_may_diverge():
    may_diverge


def test_must_diverge():
    must_diverge



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protobuf_test.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)




DESCRIPTOR = _descriptor.FileDescriptor(
  name='protobuf_test.proto',
  package='pomagma.protobuf',
  serialized_pb='\n\x13protobuf_test.proto\x12\x10pomagma.protobuf\"\xb3\x01\n\x0bTestMessage\x12\x17\n\x0foptional_string\x18\x01 \x01(\t\x12\x17\n\x0frepeated_string\x18\x16 \x03(\t\x12\x38\n\x10optional_message\x18\xcd\x02 \x01(\x0b\x32\x1d.pomagma.protobuf.TestMessage\x12\x38\n\x10repeated_message\x18\xdc\" \x03(\x0b\x32\x1d.pomagma.protobuf.TestMessage')




_TESTMESSAGE = _descriptor.Descriptor(
  name='TestMessage',
  full_name='pomagma.protobuf.TestMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='optional_string', full_name='pomagma.protobuf.TestMessage.optional_string', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='repeated_string', full_name='pomagma.protobuf.TestMessage.repeated_string', index=1,
      number=22, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='optional_message', full_name='pomagma.protobuf.TestMessage.optional_message', index=2,
      number=333, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='repeated_message', full_name='pomagma.protobuf.TestMessage.repeated_message', index=3,
      number=4444, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=42,
  serialized_end=221,
)

_TESTMESSAGE.fields_by_name['optional_message'].message_type = _TESTMESSAGE
_TESTMESSAGE.fields_by_name['repeated_message'].message_type = _TESTMESSAGE
DESCRIPTOR.message_types_by_name['TestMessage'] = _TESTMESSAGE

class TestMessage(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _TESTMESSAGE

  # @@protoc_insertion_point(class_scope:pomagma.protobuf.TestMessage)


# @@protoc_insertion_point(module_scope)



from parsable import parsable
from pomagma.util import TODO
import os
import pomagma.util

CORPUS = os.path.join(pomagma.util.SRC, 'corpus')


@parsable
def reformat(*files):
    '''
    Reformat code sections of specified markdown files in-place.
    This works recursively if given directories.
    '''
    TODO()


@parsable
def extract(filename):
    '''
    Extract code from markdown file.
    '''
    TODO()


@parsable
def annotate(*files):
    '''
    Annotate markdown files with high-latency results of analyst.
    '''
    TODO()


@parsable
def suggest(filename):
    '''
    Get real-time feedback from low-latency results of analyst.
    '''
    TODO()


@parsable
def precommit():
    '''
    Reformat and annotate all code before a git commit.
    '''
    reformat(CORPUS)


if __name__ == '__main__':
    parsable()



from pomagma import analyst
from pomagma import atlas
from pomagma import cartographer
from pomagma import linguist
from pomagma import surveyor
from pomagma import theorist
from pomagma.util import DB
from pomagma.util import suggest_region_sizes
import os
from parsable import parsable
import pomagma.io.blobstore
import pomagma.util
import pomagma.workers
import re
import shutil
import signal
import sys
import time

parsable = parsable.Parsable()

THEORY = os.environ.get('POMAGMA_THEORY', 'skrj')


# as suggested in http://stackoverflow.com/questions/974189
def raise_keyboard_interrupt(signum, frame):
    raise KeyboardInterrupt()


def already_exists(path):
    # like os.path.exists, but ignores bootstrapped worlds
    return os.path.exists(path) and not os.path.islink(path)


@parsable
def test(theory=THEORY, extra_size=0, **options):
    '''
    Test theory by building a world map.
    Options: log_level, log_file
    '''
    buildtype = 'debug' if pomagma.util.debug else 'release'
    path = os.path.join(pomagma.util.DATA, 'test', buildtype, 'atlas', theory)
    if os.path.exists(path):
        os.system('rm -f {}/*'.format(path))
    else:
        os.makedirs(path)
    with pomagma.util.chdir(path), pomagma.util.mutex(block=False):
        options.setdefault('log_file', 'test.log')
        world = DB('test.world')
        diverge_conjectures = 'diverge_conjectures.facts'
        diverge_theorems = 'diverge_theorems.facts'
        equal_conjectures = 'equal_conjectures.facts'
        equal_theorems = 'equal_theorems.facts'

        size = pomagma.util.MIN_SIZES[theory] + extra_size
        surveyor.init(theory, world, size, **options)

        with cartographer.load(theory, world, **options) as db:
            db.validate()
            for priority in [0, 1]:
                while db.infer(priority):
                    db.validate()
            db.conjecture(diverge_conjectures, equal_conjectures)
            theorist.try_prove_diverge(
                diverge_conjectures,
                diverge_conjectures,
                diverge_theorems,
                **options)
            db.assume(diverge_theorems)
            db.validate()
            db.dump(world)

        theorist.try_prove_nless(
            theory,
            world,
            equal_conjectures,
            equal_conjectures,
            equal_theorems,
            **options)

        with cartographer.load(theory, world, **options) as db:
            db.assume(equal_theorems)
            db.validate()
            for priority in [0, 1]:
                while db.infer(priority):
                    db.validate()
            db.dump(world)

        with analyst.load(theory, world, **options) as db:
            fail_count = db.test_inference()
            assert fail_count == 0, 'analyst.test_inference failed'
    print 'Theory {} appears valid.'.format(theory)


@parsable
def init(theory=THEORY, **options):
    '''
    Initialize world map for given theory.
    Options: log_level, log_file
    '''
    log_file = options.setdefault('log_file', 'init.log')
    world_size = pomagma.util.MIN_SIZES[theory]
    pomagma.util.log_print('initialize to {}'.format(world_size), log_file)
    with atlas.chdir(theory, init=True):
        survey = DB('survey')
        world = DB('world')
        normal = DB('world.normal')
        with pomagma.util.temp_copy(survey) as temp:
            surveyor.init(theory, temp, world_size, **options)
        with atlas.load(theory, survey, **options) as db:
            assert not already_exists(world), world
            assert not already_exists(normal), normal
            db.validate()
            db.dump(world)
            db.dump(normal)


@parsable
def explore(
        theory=THEORY,
        max_size=pomagma.workers.DEFAULT_SURVEY_SIZE,
        step_size=512,
        region_queue_size=4,
        **options):
    '''
    Continuously expand world map for given theory, inferring and surveying.
    Options: log_level, log_file, deadline_sec
    '''
    assert step_size > 0
    region_size = max_size - step_size
    min_size = pomagma.util.MIN_SIZES[theory]
    assert region_size >= min_size
    region_size = max(min_size, max_size - step_size)
    workers = [
        pomagma.workers.cartographer(
            theory,
            region_size,
            region_queue_size,
            **options),
        pomagma.workers.surveyor(theory, step_size, **options),
    ]
    try:
        for worker in workers:
            worker.wait()
    finally:
        for worker in workers:
            worker.terminate()


@parsable
def make(
        theory=THEORY,
        max_size=pomagma.workers.DEFAULT_SURVEY_SIZE,
        step_size=512,
        **options):
    '''
    Initialize; explore.
    Options: log_level, log_file, deadline_sec
    '''
    path = os.path.join(pomagma.util.DATA, 'atlas', theory)
    if not already_exists(path):
        init(theory, **options)
    explore(theory, max_size, step_size, **options)


@parsable
def update_theory(theory=THEORY, dry_run=False, **options):
    '''
    Update (small) world map after theory changes, and note changes.
    Options: log_level, log_file
    '''
    print dry_run
    with atlas.chdir(theory):
        world = DB('world')
        updated = pomagma.util.temp_name(DB('world'))
        assert already_exists(world), 'First initialize world map'
        options.setdefault('log_file', 'update_theory.log')
        atlas.update_theory(theory, world, updated, dry_run=dry_run, **options)


@parsable
def update_language(theory=THEORY, **options):
    '''
    Update world map after language changes.
    Options: log_level, log_file
    '''
    with atlas.chdir(theory):
        world = DB('world')
        init = pomagma.util.temp_name(DB('init'))
        aggregate = pomagma.util.temp_name(DB('aggregate'))
        assert already_exists(world), 'First initialize world map'
        options.setdefault('log_file', 'update_language.log')
        init_size = pomagma.util.MIN_SIZES[theory]
        surveyor.init(theory, init, init_size, **options)
        atlas.update_language(theory, init, world, aggregate, **options)


@parsable
def theorize(theory=THEORY, **options):
    '''
    Make conjectures based on atlas and update atlas based on theorems.
    '''
    with atlas.chdir(theory):
        world = DB('world')
        diverge_conjectures = 'diverge_conjectures.facts'
        diverge_theorems = 'diverge_theorems.facts'
        equal_conjectures = 'equal_conjectures.facts'
        nless_theorems = 'nless_theorems.facts'
        assert already_exists(world), 'First build world map'
        options.setdefault('log_file', 'theorize.log')

        with atlas.load(theory, world, **options) as db:
            db.conjecture(diverge_conjectures, equal_conjectures)
            with pomagma.util.temp_copy(diverge_conjectures) as temp:
                theorem_count = theorist.try_prove_diverge(
                    diverge_conjectures,
                    temp,
                    diverge_theorems,
                    **options)
            if theorem_count > 0:
                db.assume(diverge_theorems)
                db.dump(world)

        with pomagma.util.temp_copy(equal_conjectures) as temp:
            theorem_count = theorist.try_prove_nless(
                theory,
                world,
                equal_conjectures,
                temp,
                nless_theorems,
                **options)
        if theorem_count > 0:
            with atlas.load(theory, world, **options) as db:
                db.assume(nless_theorems)
                db.dump(world)


@parsable
def trim(theory=THEORY, parallel=True, **options):
    '''
    Trim a set of normal regions for running analyst on small machines.
    Options: log_level, log_file
    '''
    with atlas.chdir(theory):
        options.setdefault('log_file', 'trim.log')
        with cartographer.load(theory, DB('world.normal'), **options) as db:
            min_size = pomagma.util.MIN_SIZES[theory]
            max_size = db.info()['item_count']
            sizes = suggest_region_sizes(min_size, max_size)
            tasks = []
            for size in sizes:
                tasks.append({
                    'size': size,
                    'temperature': 0,
                    'filename': DB('region.normal.{:d}').format(size)
                })
            if parallel:
                print 'Trimming {} regions of sizes {}-{}'.format(
                    len(sizes),
                    sizes[0],
                    sizes[-1])
                db.trim(tasks)
            else:
                for task in tasks:
                    print 'Trimming region of size {}'.format(task['size'])
                    db.trim([task])


def _analyze(theory=THEORY, size=None, address=analyst.ADDRESS, **options):
    with atlas.chdir(theory):
        options.setdefault('log_file', 'analyst.log')
        if size is None:
            world = DB('world.normal')
        else:
            world = DB('region.normal.{}'.format(size))
        assert os.path.exists(world), 'First initialize normalized world'
        return analyst.serve(theory, world, address=address, **options)


@parsable
def analyze(theory=THEORY, size=None, address=analyst.ADDRESS, **options):
    '''
    Run analyst server on normalized world map.
    Options: log_level, log_file
    '''
    server = _analyze(theory, size, address, **options)
    try:
        server.wait()
    except KeyboardInterrupt:
        print 'stopping analyst'
    finally:
        server.stop()


@parsable
def connect(address=analyst.ADDRESS):
    '''
    Connect to analyst and start python client.
    '''
    startup = os.path.join(pomagma.util.SRC, 'analyst', 'startup.py')
    os.environ['PYTHONSTARTUP'] = startup
    os.system('python')


@parsable
def fit_language(theory, address=analyst.ADDRESS, **options):
    '''
    Fit language to corpus, saving results to git working tree.
    Options: log_level, log_file
    '''
    options.setdefault('log_file', 'linguist.log')
    linguist.fit_language(theory, address=address, **options)


match_atlas = re.compile(r'^atlas\.20\d\d(-\d\d)*').match
default_tag = time.strftime('%Y-%m-%d', time.gmtime())


def list_s3_atlases():
    import pomagma.io.s3
    filenames = pomagma.io.s3.listdir()
    return set(m.group() for m in map(match_atlas, filenames) if m)


@parsable
def pull(tag='<most recent>', force=False):
    '''
    Pull atlas from s3.
    '''
    import pomagma.io.s3
    if not os.path.exists(pomagma.util.DATA):
        os.makedirs(pomagma.util.DATA)
    with pomagma.util.chdir(pomagma.util.DATA):
        master = 'atlas'
        if os.path.exists(master):
            if force:
                shutil.rmtree(master)
            else:
                raise IOError('atlas exists; first remove atlas')
        if tag == '<most recent>':
            snapshot = max(list_s3_atlases())
        else:
            snapshot = 'atlas.{}'.format(tag)
            assert match_atlas(snapshot), 'invalid tag: {}'.format(tag)
        print 'pulling {}'.format(snapshot)
        pomagma.io.s3.pull('{}/'.format(snapshot))
        blobs = [
            os.path.join('blob', blob)
            for blob in atlas.find_used_blobs(snapshot)
        ]
        print 'pulling {} blobs'.format(len(blobs))
        pomagma.io.s3.pull(*blobs)
        pomagma.io.blobstore.validate_blobs()
        pomagma.io.s3.snapshot(snapshot, master)  # only after validation


@parsable
def push(tag=default_tag, force=False):
    '''
    Push atlas to s3.
    '''
    import pomagma.io.s3
    pomagma.io.blobstore.validate_blobs()
    with pomagma.util.chdir(pomagma.util.DATA):
        master = 'atlas'
        assert os.path.exists(master), 'atlas does not exist'
        snapshot = 'atlas.{}'.format(tag)
        assert match_atlas(snapshot), 'invalid tag: {}'.format(tag)
        if snapshot in list_s3_atlases() and not force:
            raise IOError('snapshot already exists: {}'.format(snapshot))
        print 'pushing {}'.format(snapshot)
        pomagma.io.s3.snapshot(master, snapshot)
        blobs = [
            os.path.join('blob', blob)
            for blob in atlas.find_used_blobs(snapshot)
        ]
        pomagma.io.s3.push(snapshot, *blobs)


@parsable
def gc(grace_period_days=pomagma.io.blobstore.GRACE_PERIOD_DAYS):
    '''
    Garbage collect blobs and validate remaining blobs.
    '''
    atlas.garbage_collect(grace_period_days)


if __name__ == '__main__':
    signal.signal(signal.SIGINT, raise_keyboard_interrupt)
    sys.argv[0] = 'pomagma'
    parsable()



import re
from pomagma.compiler.util import memoize_arg

re_var = re.compile('[A-Z]+$')


NARGS_TABLE = {
    'UnaryRelation': 1,
    'Equation': 2,
    'BinaryRelation': 2,
    'NullaryFunction': 0,
    'InjectiveFunction': 1,
    'BinaryFunction': 2,
    'SymmetricFunction': 2,
    'Variable': 0,
    'UnaryMeta': 1,
    'BinaryMeta': 2,
    'TernaryMeta': 3,
}

ARITY_TABLE = {
    'EQUAL': 'Equation',
    'CLOSED': 'UnaryRelation',
    'NCLOSED': 'UnaryRelation',
    'RETURN': 'UnaryRelation',
    'NRETURN': 'UnaryRelation',
    'LESS': 'BinaryRelation',
    'NLESS': 'BinaryRelation',
    'CO': 'InjectiveFunction',
    'QUOTE': 'InjectiveFunction',
    'APP': 'BinaryFunction',
    'COMP': 'BinaryFunction',
    'JOIN': 'SymmetricFunction',
    'RAND': 'SymmetricFunction',
    'VAR': 'UnaryMeta',
    'UNKNOWN': 'UnaryMeta',
    'OPTIONALLY': 'UnaryMeta',
    'NONEGATE': 'UnaryMeta',
    'EQUIVALENTLY': 'BinaryMeta',
    'FUN': 'BinaryMeta',
    'ABIND': 'TernaryMeta',
    'FIX': 'BinaryMeta',
    'FIXES': 'BinaryMeta',
}

RELATION_ARITIES = frozenset([
    'UnaryRelation',
    'Equation',
    'BinaryRelation',
])

FUNCTION_ARITIES = frozenset([
    'NullaryFunction',
    'InjectiveFunction',
    'BinaryFunction',
    'SymmetricFunction',
])

META_ARITIES = frozenset([
    'UnaryMeta',
    'BinaryMeta',
    'TernaryMeta',
])


def declare_arity(name, arity):
    assert isinstance(name, str)
    assert not is_var(name), name
    assert arity in FUNCTION_ARITIES
    if name in ARITY_TABLE:
        assert ARITY_TABLE[name] == arity, 'Cannot change arity'
    else:
        ARITY_TABLE[name] = arity


@memoize_arg
def is_var(symbol):
    return re_var.match(symbol) is None


@memoize_arg
def is_fun(symbol):
    return get_arity(symbol) in FUNCTION_ARITIES


@memoize_arg
def is_term(symbol):
    return is_var(symbol) or is_fun(symbol)


@memoize_arg
def is_rel(symbol):
    return get_arity(symbol) in RELATION_ARITIES


@memoize_arg
def is_con(symbol):
    return get_arity(symbol) in META_ARITIES


@memoize_arg
def get_arity(symbol):
    if is_var(symbol):
        return 'Variable'
    else:
        return ARITY_TABLE.get(symbol, 'NullaryFunction')


def get_nargs(arity):
    return NARGS_TABLE[arity]


def arity_sort(arity):
    return (arity in FUNCTION_ARITIES, get_nargs(arity))


def is_positive(symbol):
    if symbol == 'NLESS':
        return False
    else:
        return True


def validate():
    for symbol, arity in ARITY_TABLE.iteritems():
        assert not is_var(symbol)
        assert arity in NARGS_TABLE
    assert get_arity('x') == 'Variable'
    assert get_arity('S') == 'NullaryFunction'



import time
import multiprocessing
from nose.tools import assert_equal
import pomagma.util


def _blocking_mutex((_id, curdir)):
    with pomagma.util.chdir(curdir), pomagma.util.mutex(block=True):
        print 'Running process', _id
        time.sleep(0.1)


def test_blocking_mutex():
    THREADS = 4
    pool = multiprocessing.Pool(THREADS)
    with pomagma.util.in_temp_dir() as curdir:
        args = [(_id, curdir) for _id in xrange(THREADS)]
        pool.map(_blocking_mutex, args)


def _nonblocking_mutex((_id, curdir)):
    try:
        with pomagma.util.chdir(curdir), pomagma.util.mutex(block=False):
            print 'Running process', _id
            time.sleep(0.1)
            return 1
    except pomagma.util.MutexLockedException:
        print 'Failing process', _id
        return 0


def test_nonblocking_mutex():
    THREADS = 4
    pool = multiprocessing.Pool(THREADS)
    with pomagma.util.in_temp_dir() as curdir:
        args = [(_id, curdir) for _id in xrange(THREADS)]
        runs = pool.map(_nonblocking_mutex, args)
    assert_equal(sum(runs), 1)


def test_optimal_db_sizes():
    sizes = pomagma.util.optimal_db_sizes()
    assert_equal(sizes.next(), 1 * 512 - 1)
    assert_equal(sizes.next(), 2 * 512 - 1)
    assert_equal(sizes.next(), 3 * 512 - 1)
    assert_equal(sizes.next(), 4 * 512 - 1)
    assert_equal(sizes.next(), 6 * 512 - 1)
    assert_equal(sizes.next(), 8 * 512 - 1)
    assert_equal(sizes.next(), 12 * 512 - 1)
    assert_equal(sizes.next(), 16 * 512 - 1)


def test_suggest_region_sizes():
    sizes = pomagma.util.suggest_region_sizes(1023, 8888)
    assert_equal(sizes, [
        2 * 512 - 1,
        3 * 512 - 1,
        4 * 512 - 1,
        6 * 512 - 1,
        8 * 512 - 1,
        12 * 512 - 1,
        16 * 512 - 1,
    ])



from nose import SkipTest
from pomagma.atlas.bootstrap import THEORY
from pomagma.atlas.bootstrap import WORLD
import contextlib
import os
import pomagma.theorist.solve
import pomagma.util

DATA = os.path.join(pomagma.util.DATA, 'test', 'debug', 'atlas', THEORY)
ADDRESS = 'ipc://{}'.format(os.path.join(DATA, 'socket'))
OPTIONS = {
    'log_file': os.path.join(DATA, 'theorist_test.log'),
    'log_level': pomagma.util.LOG_LEVEL_DEBUG,
}

SKJA = os.path.join(
    pomagma.util.DATA,
    'atlas',
    'skja',
    'region.normal.{:d}.pb'.format(pomagma.util.MIN_SIZES['skja']))


@contextlib.contextmanager
def serve(world, address=ADDRESS):
    if not os.path.exists(world):
        raise SkipTest('fixture not found')
    print 'starting server'
    server = pomagma.analyst.serve(THEORY, world, address, **OPTIONS)
    yield server
    print 'stopping server'
    server.stop()


def _test_define(name):
    pomagma.theorist.solve.define(name, address=ADDRESS)


def test_define():
    with serve(WORLD):
        for name in pomagma.theorist.solve.theories:
            if name.endswith('_test'):
                yield _test_define, name
    with serve(SKJA):
        for name in pomagma.theorist.solve.theories:
            yield _test_define, name


def test_sr_pairs():
    with serve(SKJA):
        pomagma.theorist.solve.sr_pairs(address=ADDRESS)



'''
Boostrapped data is included in git repo, for testing.
'''

import os
import pomagma.util

THEORY = os.environ.get('THEORY', 'skrj')
SIZE = pomagma.util.MIN_SIZES[THEORY]
WORLD = os.path.join(
    pomagma.util.DATA,
    'atlas',
    THEORY,
    'region.normal.{:d}.pb'.format(SIZE))



from pomagma.analyst import compiler
from pomagma.analyst import messages_pb2 as messages
import sys
import zmq

CONTEXT = zmq.Context()
POLL_TIMEOUT_MS = 1000
Request = messages.AnalystRequest
Response = messages.AnalystResponse

TROOL = {
    Response.MAYBE: None,
    Response.TRUE: True,
    Response.FALSE: False,
}


def WARN(message):
    sys.stdout.write('WARNING {}\n'.format(message))
    sys.stdout.flush()


class ServerError(Exception):

    def __init__(self, messages):
        self.messages = list(messages)

    def __str__(self):
        return '\n'.join(['Server Errors:'] + self.messages)


class Client(object):

    def __init__(self, address, poll_callback=None):
        assert isinstance(address, basestring), address
        assert poll_callback is None or callable(poll_callback), poll_callback
        self._poll_callback = poll_callback
        self._socket = CONTEXT.socket(zmq.REQ)
        print 'connecting to analyst at', address
        self._socket.connect(address)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._socket.close()

    def _call(self, request):
        raw_request = request.SerializeToString()
        self._socket.send(raw_request, 0)
        if self._poll_callback is not None:
            while not self._socket.poll(timeout=POLL_TIMEOUT_MS):
                self._poll_callback()
        raw_reply = self._socket.recv(0)
        reply = Response()
        reply.ParseFromString(raw_reply)
        for message in reply.error_log:
            WARN(message)
        for key, val in request.ListFields():
            assert reply.HasField(key.name), key.name
        if reply.error_log:
            raise ServerError(reply.error_log)
        return reply

    def ping(self):
        request = Request()
        self._call(request)

    def ping_id(self, id):
        assert isinstance(id, basestring), id
        request = Request()
        request.id = id
        reply = self._call(request)
        return reply.id

    def test_inference(self):
        request = Request()
        request.test_inference.SetInParent()
        reply = self._call(request)
        return reply.test_inference.fail_count

    def _simplify(self, codes):
        request = Request()
        request.simplify.SetInParent()
        for code in codes:
            request.simplify.codes.append(compiler.desugar(code))
        reply = self._call(request)
        return list(reply.simplify.codes)

    def simplify(self, codes):
        assert isinstance(codes, list), codes
        for code in codes:
            assert isinstance(code, basestring), code
        results = self._simplify(codes)
        assert len(results) == len(codes), results
        return map(str, results)

    def _solve(self, var, theory, max_solutions):
        request = Request()
        request.solve.program = compiler.compile_solver(var, theory)
        if max_solutions is not None:
            request.solve.max_solutions = max_solutions
        reply = self._call(request)
        return {
            'necessary': map(str, reply.solve.necessary),
            'possible': map(str, reply.solve.possible),
        }

    def solve(self, var, theory, max_solutions=None):
        assert isinstance(var, basestring), var
        assert isinstance(theory, basestring), theory
        if max_solutions is not None:
            assert isinstance(max_solutions, (int, float)), max_solutions
        solutions = self._solve(var, theory, max_solutions)
        assert not (set(solutions['necessary']) & set(solutions['possible']))
        if max_solutions is not None:
            count = len(solutions['necessary']) + len(solutions['possible'])
            assert count <= max_solutions, solutions
        return solutions

    def _validate(self, codes):
        request = Request()
        request.validate.SetInParent()
        for code in codes:
            request.validate.codes.append(code)
        reply = self._call(request)
        results = []
        for result in reply.validate.results:
            results.append({
                'is_top': TROOL[result.is_top],
                'is_bot': TROOL[result.is_bot],
            })
        return results

    def validate(self, codes):
        assert isinstance(codes, list), codes
        for code in codes:
            assert isinstance(code, basestring), code
        results = self._validate(codes)
        assert len(results) == len(codes), results
        return results

    def _validate_corpus(self, lines):
        request = Request()
        request.validate_corpus.SetInParent()
        for line in lines:
            request_line = request.validate_corpus.lines.add()
            name = line['name']
            if name:
                request_line.name = name
            request_line.code = line['code']
        reply = self._call(request)
        results = []
        for result in reply.validate_corpus.results:
            results.append({
                'is_top': TROOL[result.is_top],
                'is_bot': TROOL[result.is_bot],
                'pending': result.pending,
            })
        return results

    def validate_corpus(self, lines):
        assert isinstance(lines, list), lines
        for line in lines:
            assert isinstance(line, dict), line
            assert sorted(line.keys()) == ['code', 'name']
            name = line['name']
            assert name is None or isinstance(name, basestring), name
            code = line['code']
            assert isinstance(code, basestring), code
        results = self._validate_corpus(lines)
        assert len(results) == len(lines), results
        return results

    def validate_facts(self, facts):
        assert isinstance(facts, list), facts
        request = Request()
        request.validate_facts.SetInParent()
        for fact in facts:
            assert isinstance(fact, basestring), fact
            request.validate_facts.facts.append(compiler.desugar(fact))
        reply = self._call(request)
        return TROOL[reply.validate_facts.result]

    def get_histogram(self):
        request = Request()
        request.get_histogram.SetInParent()
        reply = self._call(request)
        obs = {}
        symbols = {}
        for term in reply.get_histogram.histogram.terms:
            assert bool(term.ob) != bool(term.name), term
            count = int(term.count)
            if term.ob:
                obs[int(term.ob)] = count
            else:
                symbols[str(term.name)] = count
        result = {'obs': obs, 'symbols': symbols}
        return result

    def _fit_language(self, histogram=None):
        request = Request()
        request.fit_language.SetInParent()
        if histogram is not None:
            terms = request.fit_language.histogram.terms
            for name, count in histogram['symbols'].iteritems():
                term = terms.add()
                term.name = name
                term.count = count
            for ob, count in histogram['obs'].iteritems():
                term = terms.add()
                term.ob = ob
                term.count = count
        reply = self._call(request)
        result = {}
        for symbol in reply.fit_language.symbols:
            name = str(symbol.name)
            prob = float(symbol.prob)
            result[name] = prob
        return result

    def fit_language(self, histogram=None):
        if histogram is not None:
            assert isinstance(histogram, dict), histogram
            keys = set(histogram.keys())
            assert keys == set(['symbols', 'obs']), keys
            for name, count in histogram['symbols'].iteritems():
                assert isinstance(name, str), name
                assert isinstance(count, int), count
            for ob, count in histogram['obs'].iteritems():
                assert isinstance(ob, int), ob
                assert isinstance(count, int), count
        return self._fit_language(histogram)



from nose.tools import assert_less
from pomagma.compiler.util import eval_float44
from pomagma.compiler.util import eval_float53


def test_eval_float44():
    values = map(eval_float44, range(256))
    print ' '.join(['{}:{}'.format(k, v) for k, v in enumerate(values)])
    for i, j in zip(values[:-1], values[1:]):
        assert_less(i, j)
        assert_less(j - i - 1, 0.08 * j)   # less than 8% wasted space


def test_eval_float53():
    values = map(eval_float53, range(256))
    print ' '.join(['{}:{}'.format(k, v) for k, v in enumerate(values)])
    for i, j in zip(values[:-1], values[1:]):
        assert_less(i, j)
        assert_less(j - i - 1, 0.04 * j)   # less than 4% wasted space



from itertools import izip
from pomagma.io import creat
from pomagma.io import create_directories
import hashlib
import os
import pomagma.util
import re
import sys
import multiprocessing
import time

GRACE_PERIOD_DAYS = 7.0
RE_BLOB = re.compile('^[a-z0-9]{40}$')


def hash_file(filename):
    hasher = hashlib.sha1()
    with open(filename, 'rb') as f:
        while True:
            block = f.read(8192)
            if not block:
                break
            hasher.update(block)
    return hasher.hexdigest()


def find_blob(hexdigest):
    '''Return path to read-only file.'''
    return os.path.join(pomagma.util.BLOB_DIR, hexdigest)


def create_blob():
    '''Return temp_path to write blob to.'''
    create_directories(pomagma.util.BLOB_DIR)
    if not hasattr(create_blob, 'counter'):
        create_blob.counter = 0
    count = create_blob.counter
    create_blob.counter += 1
    filename = 'temp.{}.{}'.format(os.getpid(), count)
    path = os.path.join(pomagma.util.BLOB_DIR, filename)
    if os.path.exists(path):
        print "removing temp file ", path
        os.remove(path)
    return path


def store_blob(temp_path):
    '''Return digest for future find_blob calls; removes temp file.'''
    assert os.path.exists(temp_path)
    hexdigest = hash_file(temp_path)
    path = find_blob(hexdigest)
    if os.path.exists(path):
        os.remove(temp_path)
        os.utime(path)  # touch
    else:
        os.rename(temp_path, path)
    return hexdigest


def iter_blob_refs(filename):
    '''Iterate over all hexdigests in ref file.'''
    with open(filename, 'rb') as f:
        for line in f:
            hexdigest = line.strip()
            assert len(hexdigest) == 40, hexdigest
            yield hexdigest


def load_blob_ref(filename):
    '''Return root hexdigest from ref file.'''
    for root_hexdigest in iter_blob_refs(filename):
        return root_hexdigest


def dump_blob_ref(root_hexdigest, filename, sub_hexdigests=[]):
    '''Write root and sub hexdigests to ref file.'''
    assert isinstance(root_hexdigest, basestring), root_hexdigest
    assert len(root_hexdigest) == 40, root_hexdigest
    with creat(filename, 0444) as f:
        f.write(root_hexdigest)
        for sub_hexdigest in sub_hexdigests:
            assert isinstance(sub_hexdigest, basestring), sub_hexdigest
            assert len(sub_hexdigest) == 40, sub_hexdigest
            f.write('\n')
            f.write(sub_hexdigest)


def garbage_collect(used_blobs, grace_period_days=GRACE_PERIOD_DAYS):
    '''
    Remove all files in BLOB_DIR that:
    (1) are not reachable from refs within max_depth references; and
    (2) have not been touched within grace_period_sec.
    '''
    assert grace_period_days >= 0, grace_period_days
    grace_period_sec = grace_period_days * 24 * 3600
    for string in used_blobs:
        assert RE_BLOB.match(string), string
    used_blobs = set(used_blobs)
    deadline = time.time() - grace_period_sec
    count = 0
    for basename in os.listdir(pomagma.util.BLOB_DIR):
        if basename not in used_blobs:
            path = os.path.join(pomagma.util.BLOB_DIR, basename)
            if os.path.getmtime(path) < deadline:
                os.remove(path)
                count += 1
    print 'removed {} files from {}'.format(count, pomagma.util.BLOB_DIR)


def validate_blobs():
    '''
    Validate SHA1 and mode of some or all blobs; raise ValueError on error.
    '''
    blobs = sorted(
        blob
        for blob in os.listdir(pomagma.util.BLOB_DIR)
        if RE_BLOB.match(blob)
    )
    print 'validating {} blobs'.format(len(blobs))
    paths = [os.path.join(pomagma.util.BLOB_DIR, blob) for blob in blobs]
    hexdigests = multiprocessing.Pool().map(hash_file, paths)
    errors = [
        blob
        for blob, hexdigest in izip(blobs, hexdigests)
        if blob != hexdigest
    ]
    if errors:
        corrupt = os.path.join(pomagma.util.BLOB_DIR, 'corrupt')
        if not os.path.exists(corrupt):
            os.makedirs(corrupt)
        for error in errors:
            os.rename(
                os.path.join(pomagma.util.BLOB_DIR, error),
                os.path.join(corrupt, error))
        raise ValueError('corrupt blobs; moved to blob/corrupt/')
    for blob in blobs:
        path = os.path.join(pomagma.util.BLOB_DIR, blob)
        mode = oct(os.stat(path).st_mode)[-3:]
        if mode != '444':
            sys.stderr.write('WARNING repairing mode of {}\n'.format(blob))
            os.chmod(path, 0444)



import itertools
from pomagma.compiler.expressions import Expression
from pomagma.compiler.expressions import Expression_0
from pomagma.compiler.expressions import Expression_2
from pomagma.compiler.sequents import Sequent
from pomagma.compiler.signature import is_positive
from pomagma.compiler.util import inputs
from pomagma.compiler.util import logger
from pomagma.compiler.util import methodof
from pomagma.util import TODO

I = Expression_0('I')
K = Expression_0('K')
F = Expression_0('F')
J = Expression_0('J')
R = Expression_0('R')
B = Expression_0('B')
CB = Expression_0('CB')
C = Expression_0('C')
W = Expression_0('W')
S = Expression_0('S')
A = Expression_0('A')
BOT = Expression_0('BOT')
TOP = Expression_0('TOP')
APP = Expression_2('APP')
COMP = Expression_2('COMP')
JOIN = Expression_2('JOIN')
RAND = Expression_2('RAND')


class AbstractionFailed(Exception):
    pass


def abstract_symmetric(self, var, atom, operation):
    # K-compose-eta abstraction
    lhs, rhs = self.args
    if var in lhs.vars:
        if var in rhs.vars:
            return operation(lhs.abstract(var), rhs.abstract(var))
        elif lhs == var:
            return APP(atom, rhs)
        else:
            return COMP(APP(atom, rhs), lhs.abstract(var))
    else:
        assert var in rhs.vars
        if rhs == var:
            return APP(atom, lhs)
        else:
            return COMP(APP(atom, lhs), rhs.abstract(var))


@methodof(Expression)
def abstract(self, var):
    assert isinstance(var, Expression)
    assert var.is_var()
    if self.name == 'VAR':
        self = self.args[0]
    if self.is_var():
        if self == var:
            return I
        else:
            return APP(K, self)
    elif self.is_fun():
        name = self.name
        if var not in self.vars:
            return APP(K, self)
        elif name == 'APP':
            # I,K,C,SW,COMP,eta abstraction
            lhs, rhs = self.args
            if var in lhs.vars:
                lhs_abs = lhs.abstract(var)
                if var in rhs.vars:
                    if rhs == var:
                        return APP(W, lhs_abs)
                    else:
                        return APP(APP(S, lhs_abs), rhs.abstract(var))
                else:
                    return APP(APP(C, lhs_abs), rhs)
            else:
                assert var in rhs.vars
                if rhs == var:
                    return lhs
                else:
                    return COMP(lhs, rhs.abstract(var))
        elif name == 'COMP':
            # K,B,CB,C,S,COMP,eta abstraction
            lhs, rhs = self.args
            if var in lhs.vars:
                lhs_abs = lhs.abstract(var)
                if var in rhs.vars:
                    return APP(APP(S, COMP(B, lhs_abs)), rhs.abstract(var))
                else:
                    if lhs == var:
                        return APP(CB, rhs)
                    else:
                        return COMP(APP(CB, rhs), lhs_abs)
            else:
                assert var in rhs.vars
                if rhs == var:
                    return APP(B, lhs)
                else:
                    return COMP(APP(B, lhs), rhs.abstract(var))
        elif name == 'JOIN':
            return abstract_symmetric(self, var, J, JOIN)
        elif name == 'RAND':
            return abstract_symmetric(self, var, R, RAND)
        else:
            raise AbstractionFailed
    elif self.is_rel():
        args = [arg.abstract(var) for arg in self.args]
        return Expression.make(self.name, *args)
    else:
        raise ValueError('bad expression: %s' % self.name)


class RequireVariable(Exception):
    pass


class SkipValidation(Exception):
    pass


def get_fresh(bound):
    for name in 'abcdefghijklmnopqrstuvwxyz':
        fresh = Expression.make(name)
        if fresh not in bound:
            return fresh
    raise NotImplementedError('Exceeded fresh variable limit')


def pop_arg(args):
    if args:
        return args[0], args[1:]
    else:
        raise RequireVariable


def head_normalize(expr, *args):
    if expr.is_var():
        return [expr] + list(args)
    else:
        assert expr.is_fun(), expr
        name = expr.name
        if name == 'APP':
            lhs, rhs = expr.args
            return head_normalize(lhs, rhs, *args)
        elif name == 'COMP':
            lhs, rhs = expr.args
            arg0, args = pop_arg(args)
            return head_normalize(lhs, APP(rhs, arg0), *args)
        elif name in ['BOT', 'TOP']:
            return [expr]
        elif name == 'I':
            arg0, args = pop_arg(args)
            return head_normalize(arg0, *args)
        elif name == 'K':
            arg0, args = pop_arg(args)
            arg1, args = pop_arg(args)
            return head_normalize(arg0, *args)
        elif name == 'F':
            arg0, args = pop_arg(args)
            return head_normalize(*args)
        elif name == 'B':
            arg0, args = pop_arg(args)
            arg1, args = pop_arg(args)
            arg2, args = pop_arg(args)
            return head_normalize(arg0, APP(arg1, arg2), *args)
        elif name == 'C':
            arg0, args = pop_arg(args)
            arg1, args = pop_arg(args)
            arg2, args = pop_arg(args)
            return head_normalize(arg0, arg2, arg1, *args)
        elif name == 'W':
            arg0, args = pop_arg(args)
            arg1, args = pop_arg(args)
            return head_normalize(arg0, arg1, arg1, *args)
        elif name == 'S':
            arg0, args = pop_arg(args)
            arg1, args = pop_arg(args)
            arg2, args = pop_arg(args)
            return head_normalize(arg0, arg2, APP(arg1, arg2), *args)
        elif name == 'CI':
            return head_normalize(C, I, *args)
        elif name == 'CB':
            return head_normalize(C, B, *args)
        elif name in ['Y', 'J', 'JOIN', 'R', 'RAND', 'U', 'V', 'P', 'A']:
            raise SkipValidation
        else:
            raise TODO('head normalize %s expressions' % name)


def validate(expr):
    assert expr.is_rel(), expr
    assert expr.name in ['LESS', 'EQUAL']
    if expr.name != 'EQUAL':
        print 'WARNING: not validating {0}'.format(expr)
        return
    while True:
        try:
            lhs, rhs = expr.args
            lhs = head_normalize(lhs)
            rhs = head_normalize(rhs)
            assert len(lhs) == len(rhs),\
                'Failed to validate\n  {0}\nbecause\n  {1} != {2}'.format(
                    expr, lhs, rhs)
            assert lhs[0] == rhs[0],\
                'Failed to validate\n  {0}\nbecause  \n{1} != {2}'.format(
                    expr, lhs[0], rhs[0])
            for args in zip(lhs[1:], rhs[1:]):
                validate(Expression.make(expr.name, *args))
            break
        except RequireVariable:
            lhs, rhs = expr.args
            fresh = get_fresh(expr.vars)
            lhs = APP(lhs, fresh)
            rhs = APP(rhs, fresh)
            expr = Expression.make(expr.name, lhs, rhs)
        except SkipValidation:
            print 'WARNING: not validating {0}'.format(expr)
            return


@inputs(Expression)
def iter_eta_substitutions(expr):
    '''
    Iterate over Hindley-style substitutions:
        [x/x], [x/a], [x/APP x a] (and maybe [x/COMP x a])
    '''
    varlist = list(expr.vars)
    fresh = get_fresh(expr.vars)
    for cases in itertools.product(range(3), repeat=len(varlist)):
        result = expr
        for var, case in zip(varlist, cases):
            if case == 0:
                'do nothing'
            elif case == 1:
                result = result.substitute(var, fresh)
            elif case == 2:
                result = result.substitute(var, APP(var, fresh))
            # elif case == 3:
            #    result = result.substitute(var, COMP(var, fresh))
        if any(cases):
            yield result.abstract(fresh)
        else:
            yield result
    raise StopIteration


def iter_subsets(set_):
    list_ = list(set_)
    for cases in itertools.product([0, 1], repeat=len(list_)):
        yield set(x for (x, case) in itertools.izip(list_, cases) if case)
    raise StopIteration


@inputs(Expression)
def iter_closure_maps(expr):
    '''
    Iterate over all closing abstractions, including variable coincidence
    '''
    if not expr.vars:
        yield expr
    else:
        for varset in iter_subsets(expr.vars):
            if varset:
                var = varset.pop()
                abstracted = expr
                for other in varset:
                    abstracted = abstracted.substitute(other, var)
                abstracted = abstracted.abstract(var)
                for result in iter_closure_maps(abstracted):
                    yield result
    raise StopIteration


@inputs(Expression)
def iter_closure_permutations(expr):
    '''
    Iterate over all closing permutations
    '''
    if not expr.vars:
        yield expr
    else:
        for var in expr.vars:
            abstracted = expr.abstract(var)
            for result in iter_closure_maps(abstracted):
                yield result
    raise StopIteration


@inputs(Expression)
def iter_closures(expr):
    if expr.name in ['OPTIONALLY', 'NONEGATE']:
        expr = expr.args[0]
    try:
        assert expr.is_rel(), expr
        lhs, rhs = expr.args
        if not expr.vars:
            yield expr
        elif is_positive(expr.name):
            for expr2 in iter_eta_substitutions(expr):
                assert expr2.is_rel(), expr2
                for expr3 in iter_closure_maps(expr2):
                    assert expr3.is_rel(), expr3
                    yield expr3
        else:
            for expr2 in iter_closure_permutations(expr):
                yield expr2
    except AbstractionFailed:
        pass
    raise StopIteration


@inputs(Sequent)
def derive_facts(rule):
    facts = set()
    if len(rule.antecedents) == 0 and len(rule.succedents) == 1:
        expr = iter(rule.succedents).next()
        for derived in iter_closures(expr):
            lhs, rhs = derived.args
            if lhs != rhs:
                assert derived.is_rel()
                facts.add(derived)
        facts = sorted(list(facts), key=lambda expr: len(expr.polish))
        logger('derived {0} facts from {1}'.format(len(facts), expr))
    return facts



from pomagma.compiler.expressions import Expression_0
from pomagma.compiler.expressions import Expression_2
from pomagma.compiler.expressions import NotNegatable
from pomagma.compiler.expressions import Expression
from pomagma.compiler.expressions import try_get_negated
from pomagma.compiler.util import function
from pomagma.compiler.util import inputs
from pomagma.compiler.util import set_with
from pomagma.compiler.util import set_without
from pomagma.compiler.util import sortedset
from pomagma.compiler.util import union

BOT = Expression_0('BOT')
TOP = Expression_0('TOP')
I = Expression_0('I')
K = Expression_0('K')
F = Expression_0('F')
J = Expression_0('J')
EQUAL = Expression_2('EQUAL')
LESS = Expression_2('LESS')
NLESS = Expression_2('NLESS')

# this will be completed below
basic_facts = set([
    NLESS(TOP, BOT),
    NLESS(I, BOT), NLESS(TOP, I),
    NLESS(K, BOT), NLESS(TOP, K), NLESS(K, F),
    NLESS(F, BOT), NLESS(TOP, F), NLESS(F, K),
    NLESS(J, BOT), NLESS(TOP, J),
    LESS(K, J), NLESS(J, K),
    LESS(F, J), NLESS(J, F),
    NLESS(I, K), NLESS(K, I),
    NLESS(I, F), NLESS(F, I),
    NLESS(I, J), NLESS(J, I),
])


# TODO compile this from *.rules, rather than hand-coding
@inputs(set)
def complete_step(terms, relevant_facts):

    def step(facts):
        result = relevant_facts | facts
        equal = [p for p in result if p.name == 'EQUAL']
        less = [p for p in result if p.name == 'LESS']
        nless = [p for p in result if p.name == 'NLESS']
        # |- EQUAL x x
        # |- LESS x x
        # |- LESS x TOP
        # |- LESS BOT x
        for x in terms:
            result.add(EQUAL(x, x))
            result.add(LESS(x, x))
            result.add(LESS(BOT, x))
            result.add(LESS(x, TOP))
        # EQUAL x y |- LESS x y
        # EQUAL x y |- LESS y x
        for p in equal:
            x, y = p.args
            result.add(LESS(x, y))
            result.add(LESS(y, x))
        # LESS x y, LESS y x |- EQUAL x y
        for p in less:
            x, y = p.args
            q = LESS(y, x)
            if q in result:
                result.add(EQUAL(x, y))
        # LESS x y, LESS y z |- LESS x z
        for p in less:
            x, y = p.args
            for q in less:
                if q.args[0] == y:
                    y, z = q.args
                    result.add(LESS(x, z))
        # LESS x y, NLESS x z |- NLESS y z
        for p in less:
            x, y = p.args
            for q in nless:
                if q.args[0] == x:
                    x, z = q.args
                    result.add(NLESS(y, z))
        # NLESS x z, LESS y z |- NLESS x y
        for p in nless:
            x, z = p.args
            for q in less:
                if q.args[1] == z:
                    y, z = q.args
                    result.add(NLESS(x, y))
        return result

    return step


@inputs((set, frozenset, sortedset), function)
def close_under(facts, closure_op):
    closed = set()
    boundary = facts
    while boundary:
        closed |= boundary
        boundary = closure_op(closed)
        boundary -= closed
    return closed


def complete(facts, terms=set()):
    terms = terms | union(p.terms for p in facts) | set([BOT, TOP])
    relevant_facts = set(p for p in basic_facts if p.terms <= terms)
    return close_under(facts, complete_step(terms, relevant_facts))


basic_facts = complete(basic_facts)


@inputs(set)
def all_consistent(completed):
    negated = set()
    for p in completed:
        try:
            negated.update(try_get_negated(p))
        except NotNegatable:
            pass
    return negated.isdisjoint(completed)


class Inconsistent(Exception):
    pass


def weaken_facts(facts, fact):
    assert fact in facts
    if fact.is_rel():
        without = set_without(facts, fact)
        yield without
        if fact.name == 'EQUAL':
            lhs, rhs = fact.args
            yield set_with(without, LESS(lhs, rhs))
            yield set_with(without, LESS(rhs, lhs))


def simplify_step(completed):
    terms = union(p.terms for p in completed)

    def step(fact_sets):
        result = fact_sets.copy()
        for facts in fact_sets:
            for fact in facts:
                for simple in weaken_facts(facts, fact):
                    frozen = frozenset(simple)
                    if frozen not in result:
                        if complete(simple, terms) == completed:
                            result.add(frozen)
        return result

    return step


def objective_function(facts):
    weight = sum(2 if p.name == 'EQUAL' else 1 for p in facts if p.is_rel())
    string = ', '.join(sorted(str(p) for p in facts))
    return weight, len(string), string


@inputs(set)
def try_simplify_antecedents(facts):
    completed = complete(facts)
    if not all_consistent(completed):
        raise Inconsistent
    simple = set([frozenset(facts)])
    simple = close_under(simple, simplify_step(completed))
    facts = set(min(simple, key=objective_function))
    return facts


def strengthen_sequent(fact):
    while fact.arity == 'UnaryMeta':
        fact = fact.args[0]
    assert fact.is_rel(), fact
    if fact.name == 'LESS':
        lhs, rhs = fact.args
        if lhs == TOP or rhs == BOT:
            return EQUAL(lhs, rhs)
    return fact


def weaken_sequent(fact):
    if fact.arity == 'UnaryMeta':
        return Expression.make(fact.name, weaken_sequent(fact.args[0]))
    else:
        assert fact.is_rel(), fact
        if fact.name == 'EQUAL':
            lhs, rhs = fact.args
            if lhs == TOP or rhs == BOT:
                return LESS(lhs, rhs)
            elif rhs == TOP or lhs == BOT:
                return LESS(rhs, lhs)
        return fact
